=== added file 'utilities/ec2-generate-windmill-image.py'
--- utilities/ec2-generate-windmill-image.py	1970-01-01 00:00:00 +0000
+++ utilities/ec2-generate-windmill-image.py	2009-08-31 05:15:30 +0000
@@ -0,0 +1,712 @@
+#!/usr/bin/python
+#
+# Copyright 2009 Canonical Ltd.  This software is licensed under the
+# GNU Affero General Public License version 3 (see the file LICENSE).
+
+"""
+Generate an EC2 image that is capable of running the Windmill browser UI
+testing tool.
+
+You must provide a base image that will be augmented with the necessary
+packages and configuration.
+
+The script requires certain options to be specified in order to function
+properly.  These options may be supplied using command-line switches, or
+via a config file, with the --config command-line switch.  The default
+config file location is ~/.ec2/ec2bundle.cfg
+
+The config file format simply replicates the required command-line options
+as configuration keys.
+
+---- ec2bundle.cfg ---
+
+[DEFAULT]
+key = gsg-keypair
+identity-file = ~/.ec2/foo-keypair-id_rsa
+private-key = ~/.ec2/pk-HKZYKTAIG2ECMXYIBH3HXV4ZBZQ55CLO.pem
+cert =~/.ec2/cert-HKZYKTAIG2ECMXYIBH3HXV4ZBZQ55CLO.pem
+user-id = AIDADH4IGTRXXKCD
+access-key = SOMEBIGSTRINGOFDIGITS
+secret-key = s0m3funKyStr1Ng0fD1gitZ
+#bucket = foo  # Required, but you probably want to customize it each time.
+
+---- fin ---
+
+"""
+
+__metatype__ = type
+
+
+# Reuse a whole bunch of code from ec2test.py.
+import ConfigParser
+import ec2test
+import logging
+import optparse
+import os
+import paramiko
+import select
+import socket
+import subprocess
+import sys
+import time
+
+
+log   = logging.getLogger(__name__)
+info  = log.info
+debug = log.debug
+
+
+usage = """
+Generate an EC2 image for Windmill testing in Firefox.
+
+usage: %prog [options] AMI-ID
+"""
+
+class Instance:
+    """An EC2 instance controller."""
+
+    def __init__(self, instance):
+        self._instance = instance
+
+    @property
+    def id(self):
+        return self._instance.id
+
+    @property
+    def hostname(self):
+        return self._instance.public_dns_name
+
+    def stop(self):
+        instance = self._instance
+
+        instance.update()
+        if instance.state not in ('shutting-down', 'terminated'):
+            # terminate instance
+            instance.stop()
+            instance.update()
+
+        info('instance %s\n' % (instance.state,))
+
+    def wait_for_instance_to_start(self):
+        """Wait for the instance to transition to the "running" state."""
+        instance = self._instance
+        info('Instance %s starting..' % instance.id)
+
+        start = time.time()
+        while instance.state == 'pending':
+            sys.stdout.write('.')
+            sys.stdout.flush()
+            time.sleep(5)
+            instance.update()
+        if instance.state == 'running':
+            info('\ninstance now running at %s\n' % instance.public_dns_name)
+            elapsed = time.time() - start
+            info('Started in %d minutes %d seconds\n' %
+                     (elapsed // 60, elapsed % 60))
+            cout = instance.get_console_output()
+            info(cout.output)
+        else:
+            raise RuntimeError('failed to start: %s\n' % instance.state)
+
+    @classmethod
+    def from_image(cls, account, ami_id, instance_type):
+        """Return a new instance using the given startup parameters."""
+        info("Starting instance")
+
+        # Set up a security group that opens up ports 22, 80, and 443.  Also
+        # opens up access for our IP.
+        account.acquire_security_group()
+
+        image = account.acquire_image(ami_id)
+        key = account.name
+        debug("Image: %s, Type: %s, Key: %s" % (
+              ami_id, instance_type, key))
+
+        reservation = image.run(
+            key_name=key,
+            security_groups=[key],
+            instance_type=instance_type)
+
+        instance = cls(reservation.instances[0])
+        instance.wait_for_instance_to_start()
+        return instance
+
+    @classmethod
+    def from_running_instance(cls, account, instance_id):
+        """Create an object from an already running EC2 instance."""
+        instance = account.get_instance(instance_id)
+        if not instance:
+            raise RuntimeError(
+                "Unable to connect to instance %s" % instance_id)
+
+        info("Connected to instance %s" % instance_id)
+        proxy = cls(instance)
+        # Just to be extra safe.
+        proxy.wait_for_instance_to_start()
+        return proxy
+
+
+class SSHConnector:
+    """Handle the various aspects of using an SSH connection."""
+
+    def __init__(self, hostname, user, identity_file):
+        self.hostname = hostname
+        self.user = user
+        self.identity_file = os.path.expanduser(identity_file)
+        self._client = None
+
+    def get_private_key(self):
+        """Generate a private key object for our keyfile"""
+        fp = os.path.expanduser(self.identity_file)
+        return paramiko.RSAKey.from_private_key(open(fp))
+
+    def connect(self):
+        info('Waiting for SSH to come available: %s@%s\n' % (
+             self.user, self.hostname))
+        debug("Using private key file: %s" % self.identity_file)
+
+        private_key = self.get_private_key()
+
+        for count in range(10):
+            self._client = paramiko.SSHClient()
+            self._client.set_missing_host_key_policy(ec2test.AcceptAllPolicy())
+
+            try:
+
+                self._client.connect(
+                    self.hostname,
+                    username=self.user,
+                    pkey=private_key,
+                    allow_agent=False,
+                    look_for_keys=False)
+
+            except (socket.error, paramiko.AuthenticationException), e:
+                log.warning('wait_for_connection: %r' % (e,))
+                if count < 9:
+                    time.sleep(5)
+                    info('retrying...')
+                else:
+                    raise
+            else:
+                break
+
+    def exec_command(self, remote_command, check_return=True):
+        """Execute a command on the remote server.
+
+        Raises an error if the command returns an exit status that is not
+        zero, unless the option `check_return=False' has been given.
+        """
+        info('Executing command: %s@%s %s\n' % (self.user, self.hostname, remote_command))
+
+        session = self._client.get_transport().open_session()
+        session.exec_command(remote_command)
+        session.shutdown_write()
+
+        # TODO: change this to use the logging module
+        while True:
+            select.select([session], [], [], 0.5)
+            if session.recv_ready():
+                data = session.recv(4096)
+                if data:
+                    sys.stdout.write(data)
+                    sys.stdout.flush()
+            if session.recv_stderr_ready():
+                data = session.recv_stderr(4096)
+                if data:
+                    sys.stderr.write(data)
+                    sys.stderr.flush()
+            if session.exit_status_ready():
+                break
+        session.close()
+
+        # XXX: JonathanLange 2009-05-31: If the command is killed by a signal
+        # on the remote server, the SSH protocol does not send an exit_status,
+        # it instead sends a different message with the number of the signal
+        # that killed the process. AIUI, this code will fail confusingly if
+        # that happens.
+        exit_status = session.recv_exit_status()
+        if exit_status and check_return:
+            raise RuntimeError('Command failed: %s' % (remote_command,))
+        return exit_status
+
+    def copy_to_remote(self, local_filename, remote_filename):
+        cmd = [
+            'scp',
+            '-i', self.identity_file,
+            local_filename,
+            '%s@%s:%s' % (self.user, self.hostname, remote_filename)
+            ]
+        info("Executing command: %s" % ' '.join(cmd))
+        subprocess.check_call(cmd)
+
+    def user_command(self):
+        """Return a user-friendly ssh command-line string."""
+        return "ssh -i %s %s@%s" % (
+            self.identity_file,
+            self.user,
+            self.hostname)
+
+
+class ImageBundler:
+    """Bundle an EC2 image on a remote system."""
+
+    def __init__(self, private_key, cert, account_id, target_bucket,
+                 access_key, secret_key, ssh):
+        self.private_key = os.path.expanduser(private_key)
+        self.cert = os.path.expanduser(cert)
+        self.account_id = account_id
+        self.target_bucket = target_bucket
+        self.access_key = access_key
+        self.secret_key = secret_key
+        self.ssh = ssh
+
+        # Use the instance /mnt directory by default, because it has a few
+        # hundred GB of free space to work with.
+        self._bundle_dir = os.path.join('/mnt', target_bucket)
+
+    def bundle_image(self):
+        self.configure_bundling_environment()
+        manifest = self._bundle_image()
+        self._upload_bundle(manifest)
+        self._register_image(manifest)
+
+    def remote_private_keypath(self):
+        # ALWAYS have these files in /mnt on the remote system.  Otherwise
+        # they will get bundled along with the image.
+        return os.path.join('/mnt', os.path.basename(self.private_key))
+
+    def remote_certpath(self):
+        # ALWAYS have these files in /mnt on the remote system.  Otherwise
+        # they will get bundled along with the image.
+        return os.path.join('/mnt', os.path.basename(self.cert))
+
+    def configure_bundling_environment(self):
+        """Configure what we need on the instance for bundling the image."""
+        # Send our keypair to the remote environment so that it can be used
+        # to bundle the image.
+        local_cert = os.path.abspath(self.cert)
+        local_pkey = os.path.abspath(self.private_key)
+
+        # ALWAYS copy these files into /mnt on the remote system.  Otherwise
+        # they will get bundled along with the image.
+        remote_cert = self.remote_certpath()
+        remote_pkey = self.remote_private_keypath()
+
+        # See if the files are present, and copy them over if they are not.
+        self._ensure_remote_file(remote_cert, local_cert)
+        self._ensure_remote_file(remote_pkey, local_pkey)
+
+    def _ensure_remote_file(self, remote_file, desired_file):
+        info("Checking for '%s' on the remote system" % remote_file)
+        test = 'test -f %s' % remote_file
+        exit_status = self.ssh.exec_command(test, check_return=False)
+        if bool(exit_status):
+            self.ssh.copy_to_remote(desired_file, remote_file)
+
+    def _bundle_image(self):
+        # Create the bundle in a subdirectory, to avoid spamming up /mnt.
+        self.ssh.exec_command(
+            'mkdir %s' % self._bundle_dir, check_return=False)
+
+        cmd = [
+            'ec2-bundle-vol',
+            '-d %s' % self._bundle_dir,
+            '-b',   # Set batch-mode, which doesn't use prompts.
+            '-k %s' % self.remote_private_keypath(),
+            '-c %s' % self.remote_certpath(),
+            '-u %s'  % self.account_id
+            ]
+
+        self.ssh.exec_command(' '.join(cmd))
+        # Assume that the manifest is 'image.manifest.xml', since "image" is
+        # the default prefix.
+        manifest = os.path.join(self._bundle_dir, 'image.manifest.xml')
+
+        # Best check that the manifest actually exists though.
+        test = 'test -f %s' % manifest
+        exit_status = self.ssh.exec_command(test, check_return=False)
+
+        if bool(exit_status):
+            raise RuntimeError(
+                "Failed to write the image manifest file: %s" % manifest)
+
+        return manifest
+
+    def _upload_bundle(self, manifest):
+        cmd = [
+            'ec2-upload-bundle',
+            '-b %s' % self.target_bucket,
+            '-m %s' % manifest,
+            '-a %s' % self.access_key,
+            '-s %s' % self.secret_key
+            ]
+        self.ssh.exec_command(' '.join(cmd))
+
+    def _register_image(self, manifest):
+        # This is invoked locally.
+        mfilename = os.path.basename(manifest)
+        manifest_path = os.path.join(self.target_bucket, mfilename)
+
+        env = os.environ.copy()
+        env['JAVA_HOME'] = '/usr/lib/jvm/default-java'
+        cmd = [
+            'ec2-register',
+            '--private-key=%s' % self.private_key,
+            '--cert=%s' % self.cert,
+            manifest_path
+            ]
+        info("Executing command: %s" % ' '.join(cmd))
+        subprocess.check_call(cmd, env=env)
+
+
+class XvfbSystemConfigurator:
+    """Configure a remote operating system over SSH to use the xvfb server."""
+
+    def __init__(self, ssh):
+        self.ssh = ssh
+
+    def configure_system(self):
+        """Configure the operating system with the needed packages, etc."""
+        do = self.ssh.exec_command
+
+        # Make sure we know about all the packages, and where they may be
+        # found.
+        do("apt-get -y update")
+        # Install the necessary packages
+        do("apt-get -y install xvfb firefox xfonts-base")
+
+
+class CombinedConfigParser:
+    """Store and reconcile options for both optparse and ConfigParser."""
+
+    def __init__(self, optparser, cfgparser):
+        self._optparser = optparser
+        self._cfgparser = cfgparser
+
+        # A list of required optparse options.
+        self.required_options = []
+        self.known_options = []
+
+        # Our parsed positional command-line arguments, as returned by
+        # optparse.OptionParser.parse_args()
+        self.args = None
+
+        # An optparse option.dest to 'cfg-key' mapping.
+        self._option_to_cfgkey = {}
+
+        self._parsed_cli_options = None
+        self._parsed_cfg_options = None
+
+    def __getattr__(self, name):
+        return self.get(name)
+
+    def add_option(self, *args, **kwds):
+        """Wrap the OptionParser.add_option() method, and add our options."""
+        try:
+            # We can't pass unknown kwds to make_option, or it will barf.
+            is_required = kwds.pop('required')
+        except KeyError:
+            is_required = False
+
+        option = optparse.make_option(*args, **kwds)
+        self._optparser.add_option(option)
+
+        if is_required:
+            self.add_required_option(option)
+
+        self._add_option_to_cfg_mapping(option, args)
+
+    def add_required_option(self, option):
+        """Add a required option.
+
+        Takes an optparse.Option object.
+        """
+        self.required_options.append(option)
+
+    def _add_option_to_cfg_mapping(self, option, option_constructor_args):
+        # Convert the long options into .ini keys.  Use the last long option
+        # given.
+        for switch in reversed(option_constructor_args):
+            if switch.startswith('--'):
+                # We found a '--foo' switch, so use it.  Drop the '--',
+                # because the config file doesn't use the prefixes.
+                self._option_to_cfgkey[option.dest] = switch[2:]
+
+    def error(self, message):
+        """Wrap optparse.OptionParser.error()."""
+        self._optparser.error(message)
+
+    def parse_config_file(self, filepath):
+        fp = os.path.expanduser(filepath)
+
+        if not os.path.exists(fp):
+            self.error("The config file '%s' does not exist!" % fp)
+
+        self._cfgparser.read(fp)
+        self._parsed_cfg_options = self._cfgparser.defaults()
+
+        num_opts = len(self._parsed_cfg_options)
+        debug("Loaded %d options from %s" % (num_opts, fp))
+
+    def parse_cli_args(self, argv):
+        """Wrap optparse.OptionParser.parse_args()."""
+        options, args = self._optparser.parse_args(argv)
+        self._parsed_cli_options = options
+        self.args = args
+        return (options, args)
+
+    def verify_options(self):
+        """Verify that all required options are there.
+
+        Raise an optparse.OptionParser.error() if something is missing.
+
+        Make sure you parsed the config file with parse_config_file() before
+        doing this.
+        """
+        debug("Verifying options")
+        if not self._parsed_cfg_options:
+            debug("No config file options found")
+
+        for option in self.required_options:
+            # Check for a command-line option.
+
+            option_name = option.dest
+
+            if self.get(option_name) is None:
+                self._required_option_error(option)
+            else:
+                debug("Found required option: %s" % option_name)
+
+    def _required_option_error(self, option):
+        msg = "Required option '%s' was not given (-h for help)" % str(option)
+        self.error(msg)
+
+    def get(self, name, default=None):
+        """Return the appropriate option, CLI first, CFG second."""
+        cli_name = name
+        cfg_name = self._option_to_cfgkey.get(name)
+
+        value = self._getoption(cli_name)
+
+        if value is None and cfg_name is not None:
+            # No command-line option was supplied, but we do have a config
+            # file entry with that name.
+            value = self._getcfg(cfg_name)
+
+            if value is None:
+                # No config file option was supplied either, so return the
+                # default.
+                return default
+
+        return value
+
+    def _getoption(self, key, default=None):
+        return getattr(self._parsed_cli_options, key, default)
+
+    def _getcfg(self, key, default=None):
+        return self._parsed_cfg_options.get(key, default)
+
+
+def get_credentials():
+    """Return an EC2Credentials object for accessing the webservice."""
+    # Get the AWS identifier and secret identifier.
+    return ec2test.EC2Credentials.load_from_file()
+
+
+def parse_config_file(filepath):
+    config = ConfigParser.ConfigParser()
+    config.read(filepath)
+    return config
+
+
+def parse_options(argv):
+    oparser = optparse.OptionParser(usage)
+    cparser = ConfigParser.SafeConfigParser()
+    parser = CombinedConfigParser(oparser, cparser)
+
+    # What follows are "Required options" - these must be supplied from either
+    # the command-line, or from a config file.
+    parser.add_option(
+        '-k', '--key',
+        dest="keypair_name",
+        required=True,
+        help="The name of the AWS key pair to use for launching instances.")
+
+    parser.add_option(
+        '-K', '--private-key',
+        dest="private_key",
+        required=True,
+        help="The X.509 private keyfile that will be used to sign the new "
+             "image.")
+
+    parser.add_option(
+        '-C', '--cert',
+        dest="cert",
+        required=True,
+        help="The X.509 certificate that will be used to bundle the new "
+             "image.")
+
+    parser.add_option(
+        '-i', '--identity-file',
+        dest='identity_file',
+        required=True,
+        help="The location of the RSA private key that SSH will use to "
+             "connect to the instance.")
+
+    parser.add_option(
+        '-b', '--bucket',
+        dest="bucket",
+        required=True,
+        help="The bucket that the image will be placed into.")
+
+    parser.add_option(
+        '-u', '--user-id',
+        dest="account_id",
+        required=True,
+        help="Your 12 digit AWS account ID")
+
+    parser.add_option(
+        '-a', '--access-key',
+        dest="access_key",
+        required=True,
+        help="Your AWS access key.")
+
+    parser.add_option(
+        '-s', '--secret-key',
+        dest="secret_key",
+        required=True,
+        help="Your AWS secret key.")
+
+
+    # Start our "Optional options."
+    parser.add_option(
+        '-v', '--verbose',
+        action='store_true',
+        dest='verbose',
+        default=False,
+        help="Turn on debug output.")
+
+    parser.add_option(
+        '-t', '--instance-type',
+        dest="instance_type",
+        default="m1.large",
+        help="The type of instance to be launched.  Should be the same as "
+             "the base image's required type. [default: %default]")
+
+    parser.add_option(
+        '-c', '--config',
+        dest='config',
+        default="~/.ec2/ec2bundle.cfg",
+        help="Load script options from the supplied config file. (.ini "
+             "format, see the module docstring for details.) "
+             "[default: %default]")
+
+    parser.add_option(
+        '--keepalive',
+        action='store_true',
+        dest="keepalive",
+        default=False,
+        help="Don't shut down the instance when we are done building (or "
+             "erroring out).")
+
+    parser.add_option(
+        '--no-bundle',
+        action='store_true',
+        dest='no_bundle',
+        default=False,
+        help="Don't create a bundle, just start the server and configure the "
+             "environment.")
+
+    parser.add_option(
+        '--use-instance',
+        dest='running_instance',
+        help="Use the supplied EC2 instance ID, instead of starting our own "
+             "server.  The instance will be left running.")
+
+
+    options, args = parser.parse_cli_args(argv)
+
+    # Do this ASAP
+    if options.verbose:
+        log.setLevel(logging.DEBUG)
+
+    if options.config:
+        parser.parse_config_file(options.config)
+
+    # Make sure all the required args are present.  Will error-out if
+    # something is missing.
+    parser.verify_options()
+
+    if len(args) != 2:
+        parser.error("You must provide an AMI ID that can serve as the new "
+                     "image's base.")
+
+    return parser
+
+
+def main(argv):
+    config = parse_options(argv)
+
+    credentials = get_credentials()
+    account = credentials.connect(config.keypair_name)
+
+    # Save the flag so we can change it.  This is how we enforce shutdown
+    # policies.
+    keepalive = config.keepalive
+
+    if config.running_instance:
+        # We want to keep the server alive if the user supplied their own
+        # instance.  Killing it without their consent would be cruel.
+        keepalive = True
+
+    ssh_user_command = None
+    try:
+        try:
+            instance = None
+            if config.running_instance:
+                # Connect to an already running instance.
+                instance = Instance.from_running_instance(
+                    account, config.running_instance)
+            else:
+                # Run an instance for our base image.
+                instance = Instance.from_image(
+                    account, config.args[1], config.instance_type)
+
+            ssh = SSHConnector(
+                instance.hostname, 'root', config.identity_file)
+            ssh.connect()
+            ssh_user_command = ssh.user_command()
+
+            system_configurator = XvfbSystemConfigurator(ssh)
+            system_configurator.configure_system()
+
+            if not config.no_bundle:
+                bundler = ImageBundler(
+                    config.private_key,
+                    config.cert,
+                    config.account_id,
+                    config.bucket,
+                    config.access_key,
+                    config.secret_key,
+                    ssh)
+                bundler.bundle_image()
+
+        except:
+            # Log the exception now so it doesn't interfere with or get eaten
+            # by the instance shutdown.
+            log.exception("Oops!")
+    finally:
+        if keepalive:
+            log.warning("instance %s is now running on its own" % instance.id)
+            if ssh_user_command:
+                info("You may now ssh into the instance using the following command:")
+                info("  $ %s" % ssh_user_command)
+
+            log.warning("Remember to shut the instance down when you are done!")
+        else:
+            instance.stop()
+
+
+if __name__ == '__main__':
+    logging.basicConfig()
+    main(sys.argv)

=== added file 'utilities/ec2test-remote.py'
--- utilities/ec2test-remote.py	1970-01-01 00:00:00 +0000
+++ utilities/ec2test-remote.py	2009-08-31 05:15:30 +0000
@@ -0,0 +1,513 @@
+#!/usr/bin/env python
+# Run tests in a daemon.
+#
+# Copyright 2009 Canonical Ltd.  This software is licensed under the
+# GNU Affero General Public License version 3 (see the file LICENSE).
+
+__metatype__ = type
+
+import datetime
+import optparse
+import os
+import pickle
+import re
+import shutil
+import subprocess
+import sys
+import textwrap
+import time
+import traceback
+
+import bzrlib.branch
+import bzrlib.config
+import bzrlib.email_message
+import bzrlib.errors
+import bzrlib.smtp_connection
+import bzrlib.workingtree
+
+
+class BaseTestRunner:
+
+    def __init__(self, email=None, pqm_message=None, public_branch=None,
+                 public_branch_revno=None, test_options=None):
+        self.email = email
+        self.pqm_message = pqm_message
+        self.public_branch = public_branch
+        self.public_branch_revno = public_branch_revno
+
+        # Set up the user-supplied and default testrunner options.
+        if isinstance(test_options, basestring):
+            test_options = test_options.split()
+        elif test_options is None:
+            test_options = ['-vv']
+        else:
+            # Use whatever the user passed in.
+            pass
+        self.test_options = test_options
+
+        # Configure paths.
+        self.lp_dir = os.path.join(os.path.sep, 'var', 'launchpad')
+        self.tmp_dir = os.path.join(self.lp_dir, 'tmp')
+        self.test_dir = os.path.join(self.lp_dir, 'test')
+        self.sourcecode_dir = os.path.join(self.test_dir, 'sourcecode')
+
+        # Set up logging.
+        self.logger = WebTestLogger(
+            self.test_dir,
+            self.public_branch,
+            self.public_branch_revno,
+            self.sourcecode_dir
+        )
+
+        # Daemonization options.
+        self.pid_filename = os.path.join(self.lp_dir, 'ec2test-remote.pid')
+        self.daemonized = False
+
+    def daemonize(self):
+        """Turn the testrunner into a forked daemon process."""
+        # This also writes our pidfile to disk to a specific location.  The
+        # ec2test.py --postmortem command will look there to find our PID,
+        # in order to control this process.
+        daemonize(self.pid_filename)
+        self.daemonized = True
+
+    def remove_pidfile(self):
+        if os.path.exists(self.pid_filename):
+            os.remove(self.pid_filename)
+
+    def ignore_line(self, line):
+        """Return True if the line should be excluded from the summary log.
+
+        Defaults to False.
+        """
+        return False
+
+    def build_test_command(self):
+        """Return the command that will execute the test suite.
+
+        Should return a list of command options suitable for submission to
+        subprocess.call()
+
+        Subclasses must provide their own implementation of this method.
+        """
+        raise NotImplementedError
+
+    def test(self):
+        """Run the tests, log the results.
+
+        Signals the ec2test process and cleans up the logs once all the tests
+        have completed.  If necessary, submits the branch to PQM, and mails
+        the user the test results.
+        """
+        # We need to open the log files here because earlier calls to
+        # os.fork() may have tried to close them.
+        self.logger.prepare()
+
+        out_file     = self.logger.out_file
+        summary_file = self.logger.summary_file
+        config       = bzrlib.config.GlobalConfig()
+
+        call = self.build_test_command()
+
+        try:
+            try:
+                try:
+                    popen = subprocess.Popen(
+                        call, bufsize=-1,
+                        stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
+                        cwd=self.test_dir)
+
+                    self._gather_test_output(popen, summary_file, out_file)
+
+                    # Grab the testrunner exit status
+                    result = popen.wait()
+
+                    if self.pqm_message is not None:
+                        subject = self.pqm_message.get('Subject')
+                        if result:
+                            # failure
+                            summary_file.write(
+                                '\n\n**NOT** submitted to PQM:\n%s\n' %
+                                (subject,))
+                        else:
+                            # success
+                            conn = bzrlib.smtp_connection.SMTPConnection(
+                                config)
+                            conn.send_email(self.pqm_message)
+                            summary_file.write('\n\nSUBMITTED TO PQM:\n%s\n' %
+                                               (subject,))
+                except:
+                    summary_file.write('\n\nERROR IN TESTRUNNER\n\n')
+                    traceback.print_exc(file=summary_file)
+                    result = 1
+                    raise
+            finally:
+                # It probably isn't safe to close the log files ourselves,
+                # since someone else might try to write to them later.
+                summary_file.close()
+                if self.email is not None:
+                    subject = 'Test results: %s' % (result and 'FAILURE' or 'SUCCESS')
+                    summary_file = open(self.logger.summary_filename, 'r')
+                    bzrlib.email_message.EmailMessage.send(
+                        config, self.email[0], self.email,
+                        subject, summary_file.read())
+                    summary_file.close()
+        finally:
+            # we do this at the end because this is a trigger to ec2test.py
+            # back at home that it is OK to kill the process and take control
+            # itself, if it wants to.
+            out_file.close()
+            self.logger.close_logs()
+
+    def _gather_test_output(self, test_process, summary_file, out_file):
+        """Write the testrunner output to the logs."""
+        # Only write to stdout if we are running as the foreground process.
+        echo_to_stdout = not self.daemonized
+
+        last_line = ''
+        while 1:
+            data = test_process.stdout.read(256)
+            if data:
+                out_file.write(data)
+                out_file.flush()
+                if echo_to_stdout:
+                    sys.stdout.write(data)
+                    sys.stdout.flush()
+                lines = data.split('\n')
+                lines[0] = last_line + lines[0]
+                last_line = lines.pop()
+                for line in lines:
+                    if not self.ignore_line(line):
+                        summary_file.write(line + '\n')
+                summary_file.flush()
+            else:
+                summary_file.write(last_line)
+                break
+
+
+class TestOnMergeRunner(BaseTestRunner):
+    """Executes the Launchpad test_on_merge.py test suite."""
+
+    def build_test_command(self):
+        """See BaseTestRunner.build_test_command()."""
+        command = ['bin/py', '-t', 'test_on_merge.py']
+        command.extend(self.test_options)
+        return command
+
+    # Used to filter lines in the summary log. See
+    # `BaseTestRunner.ignore_line()`.
+    ignore_line = re.compile(
+        r'( [\w\.\/\-]+( ?\([\w\.\/\-]+\))?|'
+        r'\s*Running.*|'
+        r'\d{4}\-\d{2}\-\d{2} \d{2}\:\d{2}\:\d{2} INFO.+|'
+        r'\s*Set up .+|'
+        r'\s*Tear down .*|'
+        r'  Ran \d+ tests with .+)$').match
+
+
+class JSCheckTestRunner(BaseTestRunner):
+    """Executes the Launchpad JavaScript integration test suite."""
+
+    def build_test_command(self):
+        """See BaseTestRunner.build_test_command()."""
+        # We use the xvfb server's convenience script, xvfb-run, to
+        # automagically set the display, start the command, shut down the
+        # display, and return the exit code.  (See the xvfb-run man page for
+        # details.)
+        return [
+            'xvfb-run',
+            '-s', '-screen 0 1024x768x24',
+            'make', 'jscheck']
+
+
+class WebTestLogger:
+    """Logs test output to disk and a simple web page."""
+
+    def __init__(self, test_dir, public_branch, public_branch_revno,
+                 sourcecode_dir):
+        """ Class initialiser """
+        self.test_dir = test_dir
+        self.public_branch = public_branch
+        self.public_branch_revno = public_branch_revno
+        self.sourcecode_dir = sourcecode_dir
+
+        self.www_dir = os.path.join(os.path.sep, 'var', 'www')
+        self.out_filename = os.path.join(self.www_dir, 'current_test.log')
+        self.summary_filename = os.path.join(self.www_dir, 'summary.log')
+        self.index_filename = os.path.join(self.www_dir, 'index.html')
+
+        # We will set up the preliminary bits of the web-accessible log
+        # files. "out" holds all stdout and stderr; "summary" holds filtered
+        # output; and "index" holds an index page.
+        self.out_file = None
+        self.summary_file = None
+        self.index_file = None
+
+    def open_logs(self):
+        """Open all of our log files for writing."""
+        self.out_file = open(self.out_filename, 'w')
+        self.summary_file = open(self.summary_filename, 'w')
+        self.index_file = open(self.index_filename, 'w')
+
+    def flush_logs(self):
+        """Flush all of our log file buffers."""
+        self.out_file.flush()
+        self.summary_file.flush()
+        self.index_file.flush()
+
+    def close_logs(self):
+        """Closes all of the open log file handles."""
+        self.out_file.close()
+        self.summary_file.close()
+        self.index_file.close()
+
+    def prepare(self):
+        """Prepares the log files on disk.
+
+        Writes three log files: the raw output log, the filtered "summary"
+        log file, and a HTML index page summarizing the test run paramters.
+        """
+        self.open_logs()
+
+        out_file     = self.out_file
+        summary_file = self.summary_file
+        index_file   = self.index_file
+
+        def write(msg):
+            msg += '\n'
+            summary_file.write(msg)
+            out_file.write(msg)
+        msg = 'Tests started at approximately %(now)s UTC' % {
+            'now': datetime.datetime.utcnow().strftime(
+                '%a, %d %b %Y %H:%M:%S')}
+        index_file.write(textwrap.dedent('''\
+            <html>
+              <head>
+                <title>Testing</title>
+              </head>
+              <body>
+                <h1>Testing</h1>
+                <p>%s</p>
+                <ul>
+                  <li><a href="summary.log">Summary results</a></li>
+                  <li><a href="current_test.log">Full results</a></li>
+                </ul>
+            ''' % (msg,)))
+        write(msg)
+
+        index_file.write(textwrap.dedent('''\
+            <h2>Branches Tested</h2>
+            '''))
+
+        # Describe the trunk branch.
+        branch = bzrlib.branch.Branch.open_containing(self.test_dir)[0]
+        msg = '%(trunk)s, revision %(trunk_revno)d\n' % {
+            'trunk': branch.get_parent().encode('utf-8'),
+            'trunk_revno': branch.revno()}
+        index_file.write(textwrap.dedent('''\
+            <p><strong>%s</strong></p>
+            ''' % (msg,)))
+        write(msg)
+        tree = bzrlib.workingtree.WorkingTree.open(self.test_dir)
+        parent_ids = tree.get_parent_ids()
+
+        # Describe the merged branch.
+        if len(parent_ids) == 1:
+            index_file.write('<p>(no merged branch)</p>\n')
+            write('(no merged branch)')
+        else:
+            summary = branch.repository.get_revision(parent_ids[1]).get_summary()
+            data = {'name': self.public_branch.encode('utf-8'),
+                    'revno': self.public_branch_revno,
+                    'commit': summary}
+
+            msg = '%(name)s, revision %(revno)d (commit message: %(commit)s)\n' % data
+            index_file.write(textwrap.dedent('''\
+               <p>Merged with<br />%(msg)s</p>
+               ''' % {'msg': msg}))
+            write("Merged with")
+            write(msg)
+
+        index_file.write('<dl>\n')
+        write('\nDEPENDENCY BRANCHES USED\n')
+        for name in os.listdir(self.sourcecode_dir):
+            path = os.path.join(self.sourcecode_dir, name)
+            if os.path.isdir(path):
+                try:
+                    branch = bzrlib.branch.Branch.open_containing(path)[0]
+                except bzrlib.errors.NotBranchError:
+                    continue
+                data = {'name': name,
+                        'branch': branch.get_parent(),
+                        'revno': branch.revno()}
+                write(
+                    '- %(name)s\n    %(branch)s\n    %(revno)d\n' % data)
+                index_file.write(textwrap.dedent('''\
+                    <dt>%(name)s</dt>
+                      <dd>%(branch)s</dd>
+                      <dd>%(revno)s</dd>
+                    ''' % data))
+        index_file.write(textwrap.dedent('''\
+                </dl>
+              </body>
+            </html>'''))
+        write('\n\nTEST RESULTS FOLLOW\n\n')
+        self.flush_logs()
+
+
+def daemonize(pid_filename):
+    # this seems like the sort of thing that ought to be in the
+    # standard library :-/
+    pid = os.fork()
+    if (pid == 0): # child 1
+        os.setsid()
+        pid = os.fork()
+        if (pid == 0): # child 2
+            pass # lookie, we're ready to do work in the daemon
+        else:
+            os._exit(0)
+    else:
+        # give the pidfile a chance to be written before we exit.
+        time.sleep(1)
+        os._exit(0)
+
+    # write a pidfile ASAP
+    write_pidfile(pid_filename)
+
+   # Iterate through and close all file descriptors.
+    import resource
+    maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
+    assert maxfd != resource.RLIM_INFINITY
+    for fd in range(0, maxfd):
+        try:
+            os.close(fd)
+        except OSError:
+            # we assume fd was closed
+            pass
+    os.open(os.devnull, os.O_RDWR) # this will be 0
+    os.dup2(0, 1)
+    os.dup2(0, 2)
+
+
+def write_pidfile(pid_filename):
+    """Write a pidfile for the current process."""
+    pid_file = open(pid_filename, "w")
+    pid_file.write(str(os.getpid()))
+    pid_file.close()
+
+
+if __name__ == '__main__':
+    parser = optparse.OptionParser(
+        usage="%prog [options] [-- test options]",
+        description=("Build and run tests for an instance."))
+    parser.add_option(
+        '-e', '--email', action='append', dest='email', default=None,
+        help=('Email address to which results should be mailed.  Defaults to '
+              'the email address from `bzr whoami`. May be supplied multiple '
+              'times. The first supplied email address will be used as the '
+              'From: address.'))
+    parser.add_option(
+        '-s', '--submit-pqm-message', dest='pqm_message', default=None,
+        help=('A base64-encoded pickle (string) of a pqm message '
+              '(bzrib.plugins.pqm.pqm_submit.PQMEmailMessage) to submit if '
+              'the test run is successful.'))
+    parser.add_option(
+        '--daemon', dest='daemon', default=False,
+        action='store_true', help=('Run test in background as daemon.'))
+    parser.add_option(
+        '--debug', dest='debug', default=False,
+        action='store_true',
+        help=('Drop to pdb trace as soon as possible.'))
+    parser.add_option(
+        '--shutdown', dest='shutdown', default=False,
+        action='store_true',
+        help=('Terminate (shutdown) instance after completion.'))
+    parser.add_option(
+        '--public-branch', dest='public_branch', default=None,
+        help=('The URL of the public branch being tested.'))
+    parser.add_option(
+        '--public-branch-revno', dest='public_branch_revno',
+        type="int", default=None,
+        help=('The revision number of the public branch being tested.'))
+    parser.add_option(
+        '--jscheck', dest='jscheck', default=False, action='store_true',
+        help=('Run the JavaScript integration test suite.'))
+
+    options, args = parser.parse_args()
+
+    if options.debug:
+        import pdb; pdb.set_trace()
+    if options.pqm_message is not None:
+        pqm_message = pickle.loads(
+            options.pqm_message.decode('string-escape').decode('base64'))
+    else:
+        pqm_message = None
+
+    if options.jscheck:
+        runner_type = JSCheckTestRunner
+    else:
+        # Use the default testrunner.
+        runner_type = TestOnMergeRunner
+
+    runner = runner_type(
+       options.email,
+       pqm_message,
+       options.public_branch,
+       options.public_branch_revno,
+       args
+    )
+
+    try:
+        try:
+            if options.daemon:
+                print 'Starting testrunner daemon...'
+                runner.daemonize()
+
+            runner.test()
+        except:
+            # Handle exceptions thrown by the test() or daemonize() methods.
+            if options.email:
+                bzrlib.email_message.EmailMessage.send(
+                    bzrlib.config.GlobalConfig(), options.email[0],
+                    options.email,
+                    'Test Runner FAILED', traceback.format_exc())
+            raise
+    finally:
+
+        # When everything is over, if we've been ask to shut down, then
+        # make sure we're daemonized, then shutdown.  Otherwise, if we're
+        # daemonized, just clean up the pidfile.
+        if options.shutdown:
+            # Make sure our process is daemonized, and has therefore
+            # disconnected the controlling terminal.  This also disconnects
+            # the ec2test.py SSH connection, thus signalling ec2test.py
+            # that it may now try to take control of the server.
+            if not runner.daemonized:
+                # We only want to do this if we haven't already been
+                # daemonized.  Nesting daemons is bad.
+                runner.daemonize()
+
+            # Give the script 60 seconds to clean itself up, and 60 seconds
+            # for the ec2test.py --postmortem option to grab control if
+            # needed.  If we don't give --postmortem enough time to log
+            # in via SSH and take control, then this server will begin to
+            # shutdown on it's own.
+            #
+            # (FWIW, "grab control" means sending SIGTERM to this script's
+            # process id, thus preventing fail-safe shutdown.)
+            time.sleep(60)
+
+            # We'll only get here if --postmortem didn't kill us.  This is
+            # our fail-safe shutdown, in case the user got disconnected
+            # or suffered some other mishap that would prevent them from
+            # shutting down this server on their own.
+            subprocess.call(['sudo', 'shutdown', '-P', 'now'])
+        elif runner.daemonized:
+            # It would be nice to clean up after ourselves, since we won't
+            # be shutting down.
+            runner.remove_pidfile()
+        else:
+            # We're not a daemon, and we're not shutting down.  The user most
+            # likely started this script manually, from a shell running on the
+            # instance itself.
+            pass

=== added file 'utilities/ec2test-share-images.py'
--- utilities/ec2test-share-images.py	1970-01-01 00:00:00 +0000
+++ utilities/ec2test-share-images.py	2009-08-31 05:15:30 +0000
@@ -0,0 +1,69 @@
+#!/usr/bin/env python
+#
+# Copyright 2009 Canonical Ltd.  This software is licensed under the
+# GNU Affero General Public License version 3 (see the file LICENSE).
+
+import os
+import pprint
+import re
+import sys
+
+import boto
+
+owners = {
+    'gary': 255383312499,
+    'francis': 559320013529}
+
+_image_match = re.compile(
+    r'launchpad-ec2test(\d+)/image.manifest.xml$').match
+
+if __name__ == '__main__':
+    if len(sys.argv) == 1:
+        owner = os.environ['USER']
+    elif len(sys.argv) == 2:
+        owner = sys.argv[1]
+    else:
+        raise RuntimeError('Too many arguments')
+    try:
+        owner = int(owner)
+    except ValueError:
+        owner = owners[owner]
+    # Get the AWS identifier and secret identifier.
+    aws_id = os.path.join(os.environ['HOME'], '.ec2', 'aws_id')
+    if not os.path.exists(aws_id):
+        raise RuntimeError(
+            "Please put your aws access key identifier and secret access "
+            "key identifier in %s. (On two lines).\n" % (aws_id,))
+    aws_file = open(aws_id, 'r')
+    try:
+        identifier = aws_file.readline().strip()
+        secret = aws_file.readline().strip()
+    finally:
+        aws_file.close()
+    # Make the EC2 connection.
+    conn = boto.connect_ec2(identifier, secret)
+    # get images
+    images = {}
+    for image in conn.get_all_images(owners=owners.values()):
+        match = _image_match(image.location)
+        if match:
+            val = int(match.group(1))
+            if val not in images:
+                images[val] = [image]
+            else:
+                images[val].append(image)
+    images = images.items()
+    images.sort()
+    # set
+    last, recent = images[-2:]
+    for tmp in (last, recent):
+        if len(tmp[1]) > 1:
+            raise ValueError(
+                'more than one image of value %d found: %r' % tmp)
+    recent = recent[1][0]
+    last = last[1][0]
+    perms = last.get_launch_permissions()
+    perms['user_ids'].extend(owners.values())
+    recent.set_launch_permissions(**perms)
+    # done
+    pprint.pprint(recent.get_launch_permissions())

=== added file 'utilities/ec2test.py'
--- utilities/ec2test.py	1970-01-01 00:00:00 +0000
+++ utilities/ec2test.py	2009-08-31 05:15:30 +0000
@@ -0,0 +1,1610 @@
+#!/usr/bin/python
+# Run tests on a branch in an EC2 instance.
+#
+# Copyright 2009 Canonical Ltd.  This software is licensed under the
+# GNU Affero General Public License version 3 (see the file LICENSE).
+
+__metatype__ = type
+
+import cStringIO
+import code
+import optparse
+import os
+import pickle
+import re
+import select
+import socket
+import subprocess
+import sys
+import time
+import urllib
+import traceback
+# The rlcompleter and readline modules change the behavior of the python
+# interactive interpreter just by being imported.
+import readline
+import rlcompleter
+# Shut up pyflakes.
+rlcompleter
+
+import boto
+from boto.exception import EC2ResponseError
+from bzrlib.branch import Branch
+from bzrlib.bzrdir import BzrDir
+from bzrlib.config import GlobalConfig
+from bzrlib.errors import UncommittedChanges
+from bzrlib.plugins.launchpad.account import get_lp_login
+from bzrlib.plugins.pqm.pqm_submit import (
+    NoPQMSubmissionAddress, PQMSubmission)
+import paramiko
+
+
+TRUNK_BRANCH = 'bzr+ssh://bazaar.launchpad.net/~launchpad-pqm/launchpad/devel'
+DEFAULT_INSTANCE_TYPE = 'c1.xlarge'
+AVAILABLE_INSTANCE_TYPES = ('m1.large', 'm1.xlarge', 'c1.xlarge')
+VALID_AMI_OWNERS = (
+    255383312499, # gary
+    559320013529, # flacoste
+    200337130613, # mwhudson
+    # ...anyone else want in on the fun?
+    )
+
+readline.parse_and_bind('tab: complete')
+
+#############################################################################
+# Try to guide users past support problems we've encountered before
+if boto.Version != '1.5b':
+    msg = [
+        'Your version of python-boto (%s) is not supported.' %(boto.Version,),
+        'See https://wiki.canonical.com/Launchpad/Experiments/FiveMinutesPQM '
+        'for instructions, or search for "python-boto" among the PPAs at '
+        'https://launchpad.net/~launchpad/+archive/ppa.']
+    if not re.match('/var/lib/python-support/python2.[56]/boto',
+                    os.path.dirname(boto.__file__)):
+        # Was this boto supplied by a deb?
+        msg.append('WARNING: it looks like your version of python-boto '
+                   'might not be from a debian package.  It may be masking '
+                   'the version installed by the package manager.')
+    raise RuntimeError(' '.join(msg))
+if not paramiko.__version__.startswith('1.7.4'):
+    raise RuntimeError('Your version of paramiko (%s) is not supported.  '
+                       'Please use 1.7.4.' % (paramiko.__version__,))
+# maybe add similar check for bzrlib?
+# End
+#############################################################################
+
+#############################################################################
+# Modified from paramiko.config.  The change should be pushed upstream.
+# Our fork supports Host lines with more than one host.
+
+import fnmatch
+
+
+class SSHConfig (object):
+    """
+    Representation of config information as stored in the format used by
+    OpenSSH.  Queries can be made via L{lookup}.  The format is described in
+    OpenSSH's C{ssh_config} man page.  This class is provided primarily as a
+    convenience to posix users (since the OpenSSH format is a de-facto
+    standard on posix) but should work fine on Windows too.
+
+    @since: 1.6
+    """
+
+    def __init__(self):
+        """
+        Create a new OpenSSH config object.
+        """
+        self._config = [ { 'host': '*' } ]
+
+    def parse(self, file_obj):
+        """
+        Read an OpenSSH config from the given file object.
+
+        @param file_obj: a file-like object to read the config file from
+        @type file_obj: file
+        """
+        configs = [self._config[0]]
+        for line in file_obj:
+            line = line.rstrip('\n').lstrip()
+            if (line == '') or (line[0] == '#'):
+                continue
+            if '=' in line:
+                key, value = line.split('=', 1)
+                key = key.strip().lower()
+            else:
+                # find first whitespace, and split there
+                i = 0
+                while (i < len(line)) and not line[i].isspace():
+                    i += 1
+                if i == len(line):
+                    raise Exception('Unparsable line: %r' % line)
+                key = line[:i].lower()
+                value = line[i:].lstrip()
+
+            if key == 'host':
+                del configs[:]
+                # the value may be multiple hosts, space-delimited
+                for host in value.split():
+                    # do we have a pre-existing host config to append to?
+                    matches = [c for c in self._config if c['host'] == host]
+                    if len(matches) > 0:
+                        configs.append(matches[0])
+                    else:
+                        config = { 'host': host }
+                        self._config.append(config)
+                        configs.append(config)
+            else:
+                for config in configs:
+                    config[key] = value
+
+    def lookup(self, hostname):
+        """
+        Return a dict of config options for a given hostname.
+
+        The host-matching rules of OpenSSH's C{ssh_config} man page are used,
+        which means that all configuration options from matching host
+        specifications are merged, with more specific hostmasks taking
+        precedence.  In other words, if C{"Port"} is set under C{"Host *"}
+        and also C{"Host *.example.com"}, and the lookup is for
+        C{"ssh.example.com"}, then the port entry for C{"Host *.example.com"}
+        will win out.
+
+        The keys in the returned dict are all normalized to lowercase (look for
+        C{"port"}, not C{"Port"}.  No other processing is done to the keys or
+        values.
+
+        @param hostname: the hostname to lookup
+        @type hostname: str
+        """
+        matches = [
+            x for x in self._config if fnmatch.fnmatch(hostname, x['host'])]
+        # sort in order of shortest match (usually '*') to longest
+        matches.sort(lambda x,y: cmp(len(x['host']), len(y['host'])))
+        ret = {}
+        for m in matches:
+            ret.update(m)
+        del ret['host']
+        return ret
+
+# END paramiko config fork
+#############################################################################
+
+
+def get_ip():
+    """Uses AWS checkip to obtain this machine's IP address.
+
+    Consults an external website to determine the public IP address of this
+    machine.
+
+    :return: This machine's net-visible IP address as a string.
+    """
+    return urllib.urlopen('http://checkip.amazonaws.com').read().strip()
+
+
+class CredentialsError(Exception):
+    """Raised when AWS credentials could not be loaded."""
+
+    def __init__(self, filename, extra=None):
+        message = (
+            "Please put your aws access key identifier and secret access "
+            "key identifier in %s. (On two lines)." % (filename,))
+        if extra:
+            message += extra
+        Exception.__init__(self, message)
+
+
+class EC2Credentials:
+    """Credentials for logging in to EC2."""
+
+    DEFAULT_CREDENTIALS_FILE = '~/.ec2/aws_id'
+
+    def __init__(self, identifier, secret):
+        self.identifier = identifier
+        self.secret = secret
+
+    @classmethod
+    def load_from_file(cls, filename=None):
+        """Load the EC2 credentials from 'filename'."""
+        if filename is None:
+            filename = os.path.expanduser(cls.DEFAULT_CREDENTIALS_FILE)
+        try:
+            aws_file = open(filename, 'r')
+        except (IOError, OSError), e:
+            raise CredentialsError(filename, str(e))
+        try:
+            identifier = aws_file.readline().strip()
+            secret = aws_file.readline().strip()
+        finally:
+            aws_file.close()
+        return cls(identifier, secret)
+
+    def connect(self, name):
+        """Connect to EC2 with these credentials.
+
+        :param name: ???
+        :return: An `EC2Account` connected to EC2 with these credentials.
+        """
+        conn = boto.connect_ec2(self.identifier, self.secret)
+        return EC2Account(name, conn)
+
+
+class EC2Account:
+    """An EC2 account.
+
+    You can use this to manage security groups, keys and images for an EC2
+    account.
+    """
+
+    # Used to find pre-configured Amazon images.
+    _image_match = re.compile(
+        r'launchpad-ec2test(\d+)/image.manifest.xml$').match
+
+    def __init__(self, name, connection):
+        """Construct an EC2 instance.
+
+        :param name: ???
+        :param connection: An open boto ec2 connection.
+        """
+        self.name = name
+        self.conn = connection
+
+    def log(self, msg):
+        """Log a message on stdout, flushing afterwards."""
+        # XXX: JonathanLange 2009-05-31 bug=383076: Copied from EC2TestRunner.
+        # Should change EC2Account to take a logger and use that instead of
+        # writing to stdout.
+        sys.stdout.write(msg)
+        sys.stdout.flush()
+
+    def acquire_security_group(self, demo_networks=None):
+        """Get a security group with the appropriate configuration.
+
+        "Appropriate" means configured to allow this machine to connect via
+        SSH, HTTP and HTTPS.
+
+        If a group is already configured with this name for this connection,
+        then re-use that. Otherwise, create a new security group and configure
+        it appropriately.
+
+        The name of the security group is the `EC2Account.name` attribute.
+
+        :return: A boto security group.
+        """
+        if demo_networks is None:
+            demo_networks = []
+        try:
+            group = self.conn.get_all_security_groups(self.name)[0]
+        except EC2ResponseError, e:
+            if e.code != 'InvalidGroup.NotFound':
+                raise
+        else:
+            # If an existing security group was configured, try deleting it
+            # since our external IP might have changed.
+            try:
+                group.delete()
+            except EC2ResponseError, e:
+                if e.code != 'InvalidGroup.InUse':
+                    raise
+                # Otherwise, it means that an instance is already using
+                # it, so simply re-use it. It's unlikely that our IP changed!
+                #
+                # XXX: JonathanLange 2009-06-05: If the security group exists
+                # already, verify that the current IP is permitted; if it is
+                # not, make an INFO log and add the current IP.
+                self.log("Security group already in use, so reusing.")
+                return group
+
+        security_group = self.conn.create_security_group(
+            self.name, 'Authorization to access the test runner instance.')
+        # Authorize SSH and HTTP.
+        ip = get_ip()
+        security_group.authorize('tcp', 22, 22, '%s/32' % ip)
+        security_group.authorize('tcp', 80, 80, '%s/32' % ip)
+        security_group.authorize('tcp', 443, 443, '%s/32' % ip)
+        for network in demo_networks:
+            # Add missing netmask info for single ips.
+            if '/' not in network:
+                network += '/32'
+            security_group.authorize('tcp', 80, 80, network)
+            security_group.authorize('tcp', 443, 443, network)
+        return security_group
+
+    def acquire_private_key(self):
+        """Create & return a new key pair for the test runner."""
+        key_pair = self.conn.create_key_pair(self.name)
+        return paramiko.RSAKey.from_private_key(
+            cStringIO.StringIO(key_pair.material.encode('ascii')))
+
+    def delete_previous_key_pair(self):
+        """Delete previously used keypair, if it exists."""
+        try:
+            # Only one keypair will match 'self.name' since it's a unique
+            # identifier.
+            key_pairs = self.conn.get_all_key_pairs(self.name)
+            assert len(key_pairs) == 1, (
+                "Should be only one keypair, found %d (%s)"
+                % (len(key_pairs), key_pairs))
+            key_pair = key_pairs[0]
+            key_pair.delete()
+        except EC2ResponseError, e:
+            if e.code != 'InvalidKeyPair.NotFound':
+                if e.code == 'AuthFailure':
+                    # Inserted because of previous support issue.
+                    self.log(
+                        'POSSIBLE CAUSES OF ERROR:\n'
+                        '  Did you sign up for EC2?\n'
+                        '  Did you put a credit card number in your AWS '
+                        'account?\n'
+                        'Please doublecheck before reporting a problem.\n')
+                raise
+
+    def acquire_image(self, machine_id):
+        """Get the image.
+
+        If 'machine_id' is None, then return the image with location that
+        matches `EC2Account._image_match` and has the highest revision number
+        (where revision number is the 'NN' in 'launchpad-ec2testNN').
+
+        Otherwise, just return the image with the given 'machine_id'.
+
+        :raise ValueError: if there is more than one image with the same
+            location string.
+
+        :raise RuntimeError: if we cannot find a test-runner image.
+
+        :return: A boto image.
+        """
+        if machine_id is not None:
+            # This may raise an exception. The user specified a machine_id, so
+            # they can deal with it.
+            return self.conn.get_image(machine_id)
+
+        # We are trying to find an image that has a location that matches a
+        # regex (see definition of _image_match, above). Part of that regex is
+        # expected to be an integer with the semantics of a revision number.
+        # The image location with the highest revision number is the one that
+        # should be chosen. Because AWS does not guarantee that two images
+        # cannot share a location string, we need to make sure that the search
+        # result for this image is unique, or throw an error because the
+        # choice of image is ambiguous.
+        search_results = None
+
+        # Find the images with the highest revision numbers and locations that
+        # match the regex.
+        for image in self.conn.get_all_images(owners=VALID_AMI_OWNERS):
+            match = self._image_match(image.location)
+            if match:
+                revision = int(match.group(1))
+                if (search_results is None
+                    or search_results['revision'] < revision):
+                    # Then we have our first, highest match.
+                    search_results = {'revision': revision, 'images': [image]}
+                elif search_results['revision'] == revision:
+                    # Another image that matches and is equally high.
+                    search_results['images'].append(image)
+
+        # No matching image.
+        if search_results is None:
+            raise RuntimeError(
+                "You don't have access to a test-runner image.\n"
+                "Request access and try again.\n")
+
+        # More than one matching image.
+        if len(search_results['images']) > 1:
+            raise ValueError(
+                ('more than one image of revision %(revision)d found: '
+                 '%(images)r') % search_results)
+
+        # We could put a minimum image version number check here.
+        image = search_results['images'][0]
+        self.log(
+            'Using machine image version %d\n'
+            % (search_results['revision'],))
+        return image
+
+    def get_instance(self, instance_id):
+        """Look in all of our reservations for an instance with the given ID.
+
+        Return the instance object if it exists, None otherwise.
+        """
+        # XXX mars 20090729
+        # This method is needed by the ec2-generate-windmill-image.py script,
+        # so please do not delete it.
+        #
+        # This is a strange object on which to put this method, but I did
+        # not want to break encapsulation around the self.conn attribute.
+
+        for reservation in self.conn.get_all_instances():
+            # We need to look inside each reservation for the instances
+            # themselves.
+            for instance in reservation.instances:
+                if instance.id == instance_id:
+                    return instance
+        return None
+
+
+class UnknownBranchURL(Exception):
+    """Raised when we try to parse an unrecognized branch url."""
+
+    def __init__(self, branch_url):
+        Exception.__init__(
+            self,
+            "Couldn't parse '%s', not a Launchpad branch." % (branch_url,))
+
+
+def parse_branch_url(branch_url):
+    """Given the URL of a branch, return its components in a dict."""
+    _lp_match = re.compile(
+        r'lp:\~([^/]+)/([^/]+)/([^/]+)$').match
+    _bazaar_match = re.compile(
+        r'bzr+ssh://bazaar.launchpad.net/\~([^/]+)/([^/]+)/([^/]+)$').match
+    match = _lp_match(branch_url)
+    if match is None:
+        match = _bazaar_match(branch_url)
+    if match is None:
+        raise UnknownBranchURL(branch_url)
+    owner = match.group(1)
+    product = match.group(2)
+    branch = match.group(3)
+    unique_name = '~%s/%s/%s' % (owner, product, branch)
+    url = 'bzr+ssh://bazaar.launchpad.net/%s' % (unique_name,)
+    return dict(
+        owner=owner, product=product, branch=branch, unique_name=unique_name,
+        url=url)
+
+
+def validate_file(filename):
+    """Raise an error if 'filename' is not a file we can write to."""
+    if filename is None:
+        return
+
+    check_file = filename
+    if os.path.exists(check_file):
+        if not os.path.isfile(check_file):
+            raise ValueError(
+                'file argument %s exists and is not a file' % (filename,))
+    else:
+        check_file = os.path.dirname(check_file)
+        if (not os.path.exists(check_file) or
+            not os.path.isdir(check_file)):
+            raise ValueError(
+                'file %s cannot be created.' % (filename,))
+    if not os.access(check_file, os.W_OK):
+        raise ValueError(
+            'you do not have permission to write %s' % (filename,))
+
+
+def normalize_branch_input(data):
+    """Given 'data' return a ('dest', 'src') pair.
+
+    :param data: One of::
+       - a double of (sourcecode_location, branch_url).
+         If 'sourcecode_location' is Launchpad, then 'branch_url' can
+         also be the name of a branch of launchpad owned by
+         launchpad-pqm.
+       - a singleton of (branch_url,)
+       - a singleton of (sourcecode_location,) where
+         sourcecode_location corresponds to a Launchpad upstream
+         project as well as a rocketfuel sourcecode location.
+       - a string which could populate any of the above singletons.
+
+    :return: ('dest', 'src') where 'dest' is the destination
+        sourcecode location in the rocketfuel tree and 'src' is the
+        URL of the branch to put there. The URL can be either a bzr+ssh
+        URL or the name of a branch of launchpad owned by launchpad-pqm.
+    """
+    # XXX: JonathanLange 2009-06-05: Should convert lp: URL branches to
+    # bzr+ssh:// branches.
+    if isinstance(data, basestring):
+        data = (data,)
+    if len(data) == 2:
+        # Already in dest, src format.
+        return data
+    if len(data) != 1:
+        raise ValueError(
+            'invalid argument for ``branches`` argument: %r' %
+            (data,))
+    branch_location = data[0]
+    try:
+        parsed_url = parse_branch_url(branch_location)
+    except UnknownBranchURL:
+        return branch_location, 'lp:%s' % (branch_location,)
+    return parsed_url['product'], parsed_url['url']
+
+
+def parse_specified_branches(branches):
+    """Given 'branches' from the command line, return a sanitized dict.
+
+    The dict maps sourcecode locations to branch URLs, according to the
+    rules in `normalize_branch_input`.
+    """
+    return dict(map(normalize_branch_input, branches))
+
+
+class EC2Instance:
+    """A single EC2 instance."""
+
+    # XXX: JonathanLange 2009-05-31: Make it so that we pass one of these to
+    # EC2 test runner, rather than the test runner knowing how to make one.
+    # Right now, the test runner makes one of these directly. Instead, we want
+    # to make an EC2Account and ask it for one of these instances and then
+    # pass it to the test runner on construction.
+
+    # XXX: JonathanLange 2009-05-31: Separate out demo server maybe?
+
+    # XXX: JonathanLange 2009-05-31: Possibly separate out "get an instance"
+    # and "set up instance for Launchpad testing" logic.
+
+    def __init__(self, name, image, instance_type, demo_networks, controller,
+                 vals):
+        self._name = name
+        self._image = image
+        self._controller = controller
+        self._instance_type = instance_type
+        self._demo_networks = demo_networks
+        self._boto_instance = None
+        self._vals = vals
+
+    def error_and_quit(self, msg):
+        """Print error message and exit."""
+        sys.stderr.write(msg)
+        sys.exit(1)
+
+    def log(self, msg):
+        """Log a message on stdout, flushing afterwards."""
+        # XXX: JonathanLange 2009-05-31 bug=383076: Should delete this and use
+        # Python logging module instead.
+        sys.stdout.write(msg)
+        sys.stdout.flush()
+
+    def start(self):
+        """Start the instance."""
+        if self._boto_instance is not None:
+            self.log('Instance %s already started' % self._boto_instance.id)
+            return
+        start = time.time()
+        self.private_key = self._controller.acquire_private_key()
+        self._controller.acquire_security_group(
+            demo_networks=self._demo_networks)
+        reservation = self._image.run(
+            key_name=self._name, security_groups=[self._name],
+            instance_type=self._instance_type)
+        self._boto_instance = reservation.instances[0]
+        self.log('Instance %s starting..' % self._boto_instance.id)
+        while self._boto_instance.state == 'pending':
+            self.log('.')
+            time.sleep(5)
+            self._boto_instance.update()
+        if self._boto_instance.state == 'running':
+            self.log(' started on %s\n' % self.hostname)
+            elapsed = time.time() - start
+            self.log('Started in %d minutes %d seconds\n' %
+                     (elapsed // 60, elapsed % 60))
+            self._output = self._boto_instance.get_console_output()
+            self.log(self._output.output)
+        else:
+            self.error_and_quit(
+                'failed to start: %s\n' % self._boto_instance.state)
+
+    def shutdown(self):
+        """Shut down the instance."""
+        if self._boto_instance is None:
+            self.log('no instance created\n')
+            return
+        self._boto_instance.update()
+        if self._boto_instance.state not in ('shutting-down', 'terminated'):
+            # terminate instance
+            self._boto_instance.stop()
+            self._boto_instance.update()
+        self.log('instance %s\n' % (self._boto_instance.state,))
+
+    @property
+    def hostname(self):
+        if self._boto_instance is None:
+            return None
+        return self._boto_instance.public_dns_name
+
+    def connect_as_root(self):
+        """Connect to the instance as root.
+
+        All subsequent 'perform' and 'subprocess' operations will be done with
+        root privileges.
+        """
+        # XXX: JonathanLange 2009-06-02: This state-changing method could
+        # perhaps be written as a function such as run_as_root, or as a method
+        # that returns a root connection.
+        for count in range(10):
+            self.ssh = paramiko.SSHClient()
+            self.ssh.set_missing_host_key_policy(AcceptAllPolicy())
+            self.username = 'root'
+            try:
+                self.ssh.connect(
+                    self.hostname, username='root',
+                    pkey=self.private_key,
+                    allow_agent=False, look_for_keys=False)
+            except (socket.error, paramiko.AuthenticationException), e:
+                self.log('connect_as_root: %r' % (e,))
+                if count < 9:
+                    time.sleep(5)
+                    self.log('retrying...')
+                else:
+                    raise
+            else:
+                break
+
+    def connect_as_user(self):
+        """Connect as user.
+
+        All subsequent 'perform' and 'subprocess' operations will be done with
+        user-level privileges.
+        """
+        # XXX: JonathanLange 2009-06-02: This state-changing method could
+        # perhaps be written as a function such as run_as_user, or as a method
+        # that returns a user connection.
+        #
+        # This does not have the retry logic of connect_as_root because the
+        # circumstances that make the retries necessary appear to only happen
+        # on start-up, and connect_as_root is called first.
+        self.ssh = paramiko.SSHClient()
+        self.ssh.set_missing_host_key_policy(AcceptAllPolicy())
+        self.username = self._vals['USER']
+        self.ssh.connect(self.hostname)
+
+    def perform(self, cmd, ignore_failure=False, out=None):
+        """Perform 'cmd' on server.
+
+        :param ignore_failure: If False, raise an error on non-zero exit
+            statuses.
+        :param out: A stream to write the output of the remote command to.
+        """
+        cmd = cmd % self._vals
+        self.log('%s@%s$ %s\n' % (self.username, self._boto_instance.id, cmd))
+        session = self.ssh.get_transport().open_session()
+        session.exec_command(cmd)
+        session.shutdown_write()
+        while 1:
+            select.select([session], [], [], 0.5)
+            if session.recv_ready():
+                data = session.recv(4096)
+                if data:
+                    sys.stdout.write(data)
+                    sys.stdout.flush()
+                    if out is not None:
+                        out.write(data)
+            if session.recv_stderr_ready():
+                data = session.recv_stderr(4096)
+                if data:
+                    sys.stderr.write(data)
+                    sys.stderr.flush()
+            if session.exit_status_ready():
+                break
+        session.close()
+        # XXX: JonathanLange 2009-05-31: If the command is killed by a signal
+        # on the remote server, the SSH protocol does not send an exit_status,
+        # it instead sends a different message with the number of the signal
+        # that killed the process. AIUI, this code will fail confusingly if
+        # that happens.
+        res = session.recv_exit_status()
+        if res and not ignore_failure:
+            raise RuntimeError('Command failed: %s' % (cmd,))
+        return res
+
+    def run_with_ssh_agent(self, cmd, ignore_failure=False):
+        """Run 'cmd' in a subprocess.
+
+        Use this to run commands that require local SSH credentials. For
+        example, getting private branches from Launchpad.
+        """
+        cmd = cmd % self._vals
+        self.log('%s@%s$ %s\n' % (self.username, self._boto_instance.id, cmd))
+        call = ['ssh', '-A', self.hostname,
+               '-o', 'CheckHostIP no',
+               '-o', 'StrictHostKeyChecking no',
+               '-o', 'UserKnownHostsFile ~/.ec2/known_hosts',
+               cmd]
+        res = subprocess.call(call)
+        if res and not ignore_failure:
+            raise RuntimeError('Command failed: %s' % (cmd,))
+        return res
+
+
+class EC2TestRunner:
+
+    name = 'ec2-test-runner'
+
+    message = instance = image = None
+    _running = False
+
+    def __init__(self, branch, email=False, file=None, test_options='-vv',
+                 headless=False, branches=(),
+                 machine_id=None, instance_type=DEFAULT_INSTANCE_TYPE,
+                 pqm_message=None, pqm_public_location=None,
+                 pqm_submit_location=None, demo_networks=None,
+                 open_browser=False, pqm_email=None,
+                 include_download_cache_changes=None):
+        """Create a new EC2TestRunner.
+
+        This sets the following attributes:
+          - original_branch
+          - test_options
+          - headless
+          - include_download_cache_changes
+          - download_cache_additions
+          - branches (parses, validates)
+          - message (after validating PQM submisson)
+          - email (after validating email capabilities)
+          - instance_type (validates)
+          - image (after connecting to ec2)
+          - file (after checking we can write to it)
+          - ssh_config_file_name (after checking it exists)
+          - vals, a dict containing
+            - the environment
+            - trunk_branch (either from global or derived from branches)
+            - branch
+            - smtp_server
+            - smtp_username
+            - smtp_password
+            - email (distinct from the email attribute)
+            - key_type
+            - key
+            - launchpad_login
+        """
+        self.original_branch = branch # just for easy access in debugging
+        self.test_options = test_options
+        self.headless = headless
+        self.include_download_cache_changes = include_download_cache_changes
+        if demo_networks is None:
+            demo_networks = ()
+        else:
+            demo_networks = demo_networks
+        self.open_browser = open_browser
+        if headless and file:
+            raise ValueError(
+                'currently do not support files with headless mode.')
+        if headless and not (email or pqm_message):
+            raise ValueError('You have specified no way to get the results '
+                             'of your headless test run.')
+
+        if test_options != '-vv' and pqm_message is not None:
+            raise ValueError(
+                "Submitting to PQM with non-default test options isn't "
+                "supported")
+
+        trunk_specified = False
+        trunk_branch = TRUNK_BRANCH
+
+        # normalize and validate branches
+        branches = parse_specified_branches(branches)
+        try:
+            launchpad_url = branches.pop('launchpad')
+        except KeyError:
+            # No Launchpad branch specified.
+            pass
+        else:
+            try:
+                parsed_url = parse_branch_url(launchpad_url)
+            except UnknownBranchURL:
+                user = 'launchpad-pqm'
+                src = ('bzr+ssh://bazaar.launchpad.net/'
+                       '~launchpad-pqm/launchpad/%s' % (launchpad_url,))
+            else:
+                user = parsed_url['owner']
+                src = parsed_url['url']
+            if user == 'launchpad-pqm':
+                trunk_specified = True
+            trunk_branch = src
+
+        self.branches = branches.items()
+
+        # XXX: JonathanLange 2009-05-31: The trunk_specified stuff above and
+        # the pqm location stuff below are actually doing the equivalent of
+        # preparing a merge directive. Perhaps we can leverage that to make
+        # this code simpler.
+        self.download_cache_additions = None
+        if branch is None:
+            config = GlobalConfig()
+            if pqm_message is not None:
+                raise ValueError('Cannot submit trunk to pqm.')
+        else:
+            (tree,
+             bzrbranch,
+             relpath) = BzrDir.open_containing_tree_or_branch(branch)
+            # if tree is None, remote...I'm assuming.
+            if tree is None:
+                config = GlobalConfig()
+            else:
+                config = bzrbranch.get_config()
+
+            if pqm_message is not None or tree is not None:
+                # if we are going to maybe send a pqm_message, we're going to
+                # go down this path. Also, even if we are not but this is a
+                # local branch, we're going to use the PQM machinery to make
+                # sure that the local branch has been made public, and has all
+                # working changes there.
+                if tree is None:
+                    # remote.  We will make some assumptions.
+                    if pqm_public_location is None:
+                        pqm_public_location = branch
+                    if pqm_submit_location is None:
+                        pqm_submit_location = trunk_branch
+                elif pqm_submit_location is None and trunk_specified:
+                    pqm_submit_location = trunk_branch
+                # modified from pqm_submit.py
+                submission = PQMSubmission(
+                    source_branch=bzrbranch,
+                    public_location=pqm_public_location,
+                    message=pqm_message or '',
+                    submit_location=pqm_submit_location,
+                    tree=tree)
+                if tree is not None:
+                    # this is the part we want to do whether or not we're
+                    # submitting.
+                    submission.check_tree() # any working changes
+                    submission.check_public_branch() # everything public
+                    branch = submission.public_location
+                    if (include_download_cache_changes is None or
+                        include_download_cache_changes):
+                        # We need to get the download cache settings
+                        cache_tree, cache_bzrbranch, cache_relpath = (
+                            BzrDir.open_containing_tree_or_branch(
+                                os.path.join(
+                                    self.original_branch, 'download-cache')))
+                        cache_tree.lock_read()
+                        try:
+                            cache_basis_tree = cache_tree.basis_tree()
+                            cache_basis_tree.lock_read()
+                            try:
+                                delta = cache_tree.changes_from(
+                                    cache_basis_tree, want_unversioned=True)
+                                unversioned = [
+                                    un for un in delta.unversioned
+                                    if not cache_tree.is_ignored(un[0])]
+                                added = delta.added
+                                self.download_cache_additions = (
+                                    unversioned + added)
+                            finally:
+                                cache_basis_tree.unlock()
+                        finally:
+                            cache_tree.unlock()
+                if pqm_message is not None:
+                    if self.download_cache_additions:
+                        raise UncommittedChanges(cache_tree)
+                    # get the submission message
+                    mail_from = config.get_user_option('pqm_user_email')
+                    if not mail_from:
+                        mail_from = config.username()
+                    # Make sure this isn't unicode
+                    mail_from = mail_from.encode('utf8')
+                    if pqm_email is None:
+                        if tree is None:
+                            pqm_email = (
+                                "Launchpad PQM <launchpad@pqm.canonical.com>")
+                        else:
+                            pqm_email = config.get_user_option('pqm_email')
+                    if not pqm_email:
+                        raise NoPQMSubmissionAddress(bzrbranch)
+                    mail_to = pqm_email.encode('utf8') # same here
+                    self.message = submission.to_email(mail_from, mail_to)
+                elif (self.download_cache_additions and
+                      self.include_download_cache_changes is None):
+                    raise UncommittedChanges(
+                        cache_tree,
+                        'You must select whether to include download cache '
+                        'changes (see --include-download-cache-changes and '
+                        '--ignore-download-cache-changes, -c and -g '
+                        'respectively), or '
+                        'commit or remove the files in the download-cache.')
+        if email is not False:
+            if email is True:
+                email = [config.username()]
+                if not email[0]:
+                    raise ValueError('cannot find your email address.')
+            elif isinstance(email, basestring):
+                email = [email]
+            else:
+                tmp = []
+                for item in email:
+                    if not isinstance(item, basestring):
+                        raise ValueError(
+                            'email must be True, False, a string, or a list of '
+                            'strings')
+                    tmp.append(item)
+                email = tmp
+        else:
+            email = None
+        self.email = email
+
+        # We do a lot of looking before leaping here because we want to avoid
+        # wasting time and money on errors we could have caught early.
+
+        # Validate instance_type and get default kernal and ramdisk.
+        if instance_type not in AVAILABLE_INSTANCE_TYPES:
+            raise ValueError('unknown instance_type %s' % (instance_type,))
+
+        # Validate and set file.
+        validate_file(file)
+        self.file = file
+
+        # Make a dict for string substitution based on the environ.
+        #
+        # XXX: JonathanLange 2009-06-02: Although this defintely makes the
+        # scripts & commands easier to write, it makes it harder to figure out
+        # how the different bits of the system interoperate (passing 'vals' to
+        # a method means it uses...?). Consider changing things around so that
+        # vals is not needed.
+        self.vals = dict(os.environ)
+        self.vals['trunk_branch'] = trunk_branch
+        self.vals['branch'] = branch
+        home = self.vals['HOME']
+
+        # Email configuration.
+        if email is not None or pqm_message is not None:
+            server = self.vals['smtp_server'] = config.get_user_option(
+                'smtp_server')
+            if server is None or server == 'localhost':
+                raise ValueError(
+                    'To send email, a remotely accessible smtp_server (and '
+                    'smtp_username and smtp_password, if necessary) must be '
+                    'configured in bzr.  See the SMTP server information '
+                    'here: https://wiki.canonical.com/EmailSetup .')
+            self.vals['smtp_username'] = config.get_user_option(
+                'smtp_username')
+            self.vals['smtp_password'] = config.get_user_option(
+                'smtp_password')
+            from_email = config.username()
+            if not from_email:
+                raise ValueError(
+                    'To send email, your bzr email address must be set '
+                    '(use ``bzr whoami``).')
+            else:
+                self.vals['email'] = (
+                    from_email.encode('utf8').encode('string-escape'))
+
+        # Get a public key from the agent.
+        agent = paramiko.Agent()
+        keys = agent.get_keys()
+        if len(keys) == 0:
+            self.error_and_quit(
+                'You must have an ssh agent running with keys installed that '
+                'will allow the script to rsync to devpad and get your '
+                'branch.\n')
+        key = agent.get_keys()[0]
+        self.vals['key_type'] = key.get_name()
+        self.vals['key'] = key.get_base64()
+
+        # Verify the .ssh config file
+        self.ssh_config_file_name = os.path.join(home, '.ssh', 'config')
+        if not os.path.exists(self.ssh_config_file_name):
+            self.error_and_quit(
+                'This script expects to find the .ssh config in %s.  Please '
+                'make sure it exists and contains the necessary '
+                'configuration to access devpad.' % (
+                    self.ssh_config_file_name,))
+
+        # Get the bzr login.
+        login = get_lp_login()
+        if not login:
+            self.error_and_quit(
+                'you must have set your launchpad login in bzr.')
+        self.vals['launchpad-login'] = login
+
+        # Get the AWS identifier and secret identifier.
+        try:
+            credentials = EC2Credentials.load_from_file()
+        except CredentialsError, e:
+            self.error_and_quit(str(e))
+
+        # Make the EC2 connection.
+        controller = credentials.connect(self.name)
+
+        # We do this here because it (1) cleans things up and (2) verifies
+        # that the account is correctly set up. Both of these are appropriate
+        # for initialization.
+        #
+        # We always recreate the keypairs because there is no way to
+        # programmatically retrieve the private key component, unless we
+        # generate it.
+        controller.delete_previous_key_pair()
+
+        # get the image
+        image = controller.acquire_image(machine_id)
+        self._instance = EC2Instance(
+            self.name, image, instance_type, demo_networks,
+            controller, self.vals)
+        # now, as best as we can tell, we should be good to go.
+
+    def error_and_quit(self, msg):
+        """Print error message and exit."""
+        sys.stderr.write(msg)
+        sys.exit(1)
+
+    def log(self, msg):
+        """Log a message on stdout, flushing afterwards."""
+        # XXX: JonathanLange 2009-05-31 bug=383076: This should use Python
+        # logging, rather than printing to stdout.
+        sys.stdout.write(msg)
+        sys.stdout.flush()
+
+    def start(self):
+        """Start the EC2 instance."""
+        self._instance.start()
+
+    def shutdown(self):
+        if self.headless and self._running:
+            self.log('letting instance run, to shut down headlessly '
+                     'at completion of tests.\n')
+            return
+        return self._instance.shutdown()
+
+    def configure_system(self):
+        # AS ROOT
+        self._instance.connect_as_root()
+        if self.vals['USER'] == 'gary':
+            # This helps gary debug problems others are having by removing
+            # much of the initial setup used to work on the original image.
+            self._instance.perform('deluser --remove-home gary',
+                          ignore_failure=True)
+        p = self._instance.perform
+        # Let root perform sudo without a password.
+        p('echo "root\tALL=NOPASSWD: ALL" >> /etc/sudoers')
+        # Add the user.
+        p('adduser --gecos "" --disabled-password %(USER)s')
+        # Give user sudo without password.
+        p('echo "%(USER)s\tALL=(ALL) NOPASSWD: ALL" >> /etc/sudoers')
+            # Make /var/launchpad owned by user.
+        p('chown -R %(USER)s:%(USER)s /var/launchpad')
+        # Clean out left-overs from the instance image.
+        p('rm -fr /var/tmp/*')
+        # Update the system.
+        p('aptitude update')
+        p('aptitude -y full-upgrade')
+        # Set up ssh for user
+        # Make user's .ssh directory
+        p('sudo -u %(USER)s mkdir /home/%(USER)s/.ssh')
+        sftp = self._instance.ssh.open_sftp()
+        remote_ssh_dir = '/home/%(USER)s/.ssh' % self.vals
+        # Create config file
+        self.log('Creating %s/config\n' % (remote_ssh_dir,))
+        ssh_config_source = open(self.ssh_config_file_name)
+        config = SSHConfig()
+        config.parse(ssh_config_source)
+        ssh_config_source.close()
+        ssh_config_dest = sftp.open("%s/config" % remote_ssh_dir, 'w')
+        ssh_config_dest.write('CheckHostIP no\n')
+        ssh_config_dest.write('StrictHostKeyChecking no\n')
+        for hostname in ('devpad.canonical.com', 'chinstrap.canonical.com'):
+            ssh_config_dest.write('Host %s\n' % (hostname,))
+            data = config.lookup(hostname)
+            for key in ('hostname', 'gssapiauthentication', 'proxycommand',
+                        'user', 'forwardagent'):
+                value = data.get(key)
+                if value is not None:
+                    ssh_config_dest.write('    %s %s\n' % (key, value))
+        ssh_config_dest.write('Host bazaar.launchpad.net\n')
+        ssh_config_dest.write('    user %(launchpad-login)s\n' % self.vals)
+        ssh_config_dest.close()
+        # create authorized_keys
+        self.log('Setting up %s/authorized_keys\n' % remote_ssh_dir)
+        authorized_keys_file = sftp.open(
+            "%s/authorized_keys" % remote_ssh_dir, 'w')
+        authorized_keys_file.write("%(key_type)s %(key)s\n" % self.vals)
+        authorized_keys_file.close()
+        sftp.close()
+        # Chown and chmod the .ssh directory and contents that we just
+        # created.
+        p('chown -R %(USER)s:%(USER)s /home/%(USER)s/')
+        p('chmod 644 /home/%(USER)s/.ssh/*')
+        self.log(
+            'You can now use ssh -A %s to log in the instance.\n' %
+            self._instance.hostname)
+        # give the user permission to do whatever in /var/www
+        p('chown -R %(USER)s:%(USER)s /var/www')
+        self._instance.ssh.close()
+
+        # AS USER
+        self._instance.connect_as_user()
+        sftp = self._instance.ssh.open_sftp()
+        # Set up bazaar.conf with smtp information if necessary
+        if self.email or self.message:
+            p('sudo -u %(USER)s mkdir /home/%(USER)s/.bazaar')
+            bazaar_conf_file = sftp.open(
+                "/home/%(USER)s/.bazaar/bazaar.conf" % self.vals, 'w')
+            bazaar_conf_file.write(
+                'smtp_server = %(smtp_server)s\n' % self.vals)
+            if self.vals['smtp_username']:
+                bazaar_conf_file.write(
+                    'smtp_username = %(smtp_username)s\n' % self.vals)
+            if self.vals['smtp_password']:
+                bazaar_conf_file.write(
+                    'smtp_password = %(smtp_password)s\n' % self.vals)
+            bazaar_conf_file.close()
+        # Copy remote ec2-remote over
+        self.log('Copying ec2test-remote.py to remote machine.\n')
+        sftp.put(
+            os.path.join(os.path.dirname(os.path.realpath(__file__)),
+                         'ec2test-remote.py'),
+            '/var/launchpad/ec2test-remote.py')
+        sftp.close()
+        # Set up launchpad login and email
+        p('bzr launchpad-login %(launchpad-login)s')
+        p("bzr whoami '%(email)s'")
+        self._instance.ssh.close()
+
+    def prepare_tests(self):
+        self._instance.connect_as_user()
+        # Clean up the test branch left in the instance image.
+        self._instance.perform('rm -rf /var/launchpad/test')
+        # get newest sources
+        self._instance.run_with_ssh_agent(
+            "rsync -avp --partial --delete "
+            "--filter='P *.o' --filter='P *.pyc' --filter='P *.so' "
+            "devpad.canonical.com:/code/rocketfuel-built/launchpad/sourcecode/* "
+            "/var/launchpad/sourcecode/")
+        # Get trunk.
+        self._instance.run_with_ssh_agent(
+            'bzr branch %(trunk_branch)s /var/launchpad/test')
+        # Merge the branch in.
+        if self.vals['branch'] is not None:
+            self._instance.run_with_ssh_agent(
+                'cd /var/launchpad/test; bzr merge %(branch)s')
+        else:
+            self.log('(Testing trunk, so no branch merge.)')
+        # Get any new sourcecode branches as requested
+        for dest, src in self.branches:
+            fulldest = os.path.join('/var/launchpad/test/sourcecode', dest)
+            if dest in ('canonical-identity-provider', 'shipit'):
+                # These two branches share some of the history with Launchpad.
+                # So we create a stacked branch on Launchpad so that the shared
+                # history isn't duplicated.
+                self._instance.run_with_ssh_agent(
+                    'bzr branch --no-tree --stacked %s %s' %
+                    (TRUNK_BRANCH, fulldest))
+                # The --overwrite is needed because they are actually two
+                # different branches (canonical-identity-provider was not
+                # branched off launchpad, but some revisions are shared.)
+                self._instance.run_with_ssh_agent(
+                    'bzr pull --overwrite %s -d %s' % (src, fulldest))
+                # The third line is necessary because of the --no-tree option
+                # used initially. --no-tree doesn't create a working tree.
+                # It only works with the .bzr directory (branch metadata and
+                # revisions history). The third line creates a working tree
+                # based on the actual branch.
+                self._instance.run_with_ssh_agent(
+                    'bzr checkout "%s" "%s"' % (fulldest, fulldest))
+            else:
+                # The "--standalone" option is needed because some branches
+                # are/were using a different repository format than Launchpad
+                # (bzr-svn branch for example).
+                self._instance.run_with_ssh_agent(
+                    'bzr branch --standalone %s %s' % (src, fulldest))
+        # prepare fresh copy of sourcecode and buildout sources for building
+        p = self._instance.perform
+        p('rm -rf /var/launchpad/tmp')
+        p('mkdir /var/launchpad/tmp')
+        p('cp -R /var/launchpad/sourcecode /var/launchpad/tmp/sourcecode')
+        p('mkdir /var/launchpad/tmp/eggs')
+        self._instance.run_with_ssh_agent(
+            'bzr co lp:lp-source-dependencies '
+            '/var/launchpad/tmp/download-cache')
+        if (self.include_download_cache_changes and
+            self.download_cache_additions):
+            sftp = self._instance.ssh.open_sftp()
+            root = os.path.realpath(
+                os.path.join(self.original_branch, 'download-cache'))
+            for info in self.download_cache_additions:
+                src = os.path.join(root, info[0])
+                self.log('Copying %s to remote machine.\n' % (src,))
+                sftp.put(
+                    src,
+                    os.path.join('/var/launchpad/tmp/download-cache', info[0]))
+            sftp.close()
+        p('/var/launchpad/test/utilities/link-external-sourcecode '
+          '-p/var/launchpad/tmp -t/var/launchpad/test'),
+        # set up database
+        p('/var/launchpad/test/utilities/launchpad-database-setup %(USER)s')
+        p('cd /var/launchpad/test && make build')
+        p('cd /var/launchpad/test && make schema')
+        # close ssh connection
+        self._instance.ssh.close()
+
+    def start_demo_webserver(self):
+        """Turn ec2 instance into a demo server."""
+        self._instance.connect_as_user()
+        p = self._instance.perform
+        p('mkdir -p /var/tmp/bazaar.launchpad.dev/static')
+        p('mkdir -p /var/tmp/bazaar.launchpad.dev/mirrors')
+        p('sudo a2enmod proxy > /dev/null')
+        p('sudo a2enmod proxy_http > /dev/null')
+        p('sudo a2enmod rewrite > /dev/null')
+        p('sudo a2enmod ssl > /dev/null')
+        p('sudo a2enmod deflate > /dev/null')
+        p('sudo a2enmod headers > /dev/null')
+        # Install apache config file.
+        p('cd /var/launchpad/test/; sudo make install')
+        # Use raw string to eliminate the need to escape the backslash.
+        # Put eth0's ip address in the /tmp/ip file.
+        p(r"ifconfig eth0 | grep 'inet addr' "
+          r"| sed -re 's/.*addr:([0-9.]*) .*/\1/' > /tmp/ip")
+        # Replace 127.0.0.88 in Launchpad's apache config file with the
+        # ip address just stored in the /tmp/ip file. Perl allows for
+        # inplace editing unlike sed.
+        p('sudo perl -pi -e "s/127.0.0.88/$(cat /tmp/ip)/g" '
+          '/etc/apache2/sites-available/local-launchpad')
+        # Restart apache.
+        p('sudo /etc/init.d/apache2 restart')
+        # Build mailman and minified javascript, etc.
+        p('cd /var/launchpad/test/; make')
+        # Start launchpad in the background.
+        p('cd /var/launchpad/test/; make start')
+        # close ssh connection
+        self._instance.ssh.close()
+
+    def run_tests(self):
+        self._instance.connect_as_user()
+
+        # Make sure we activate the failsafe --shutdown feature.  This will
+        # make the server shut itself down after the test run completes, or
+        # if the test harness suffers a critical failure.
+        cmd = ['python /var/launchpad/ec2test-remote.py --shutdown']
+
+        # Do we want to email the results to the user?
+        if self.email:
+            for email in self.email:
+                cmd.append("--email='%s'" % (
+                    email.encode('utf8').encode('string-escape'),))
+
+        # Do we want to submit the branch to PQM if the tests pass?
+        if self.message is not None:
+            cmd.append(
+                "--submit-pqm-message='%s'" % (
+                    pickle.dumps(
+                        self.message).encode(
+                        'base64').encode('string-escape'),))
+
+        # Do we want to disconnect the terminal once the test run starts?
+        if self.headless:
+            cmd.append('--daemon')
+
+        # Which branch do we want to test?
+        if self.vals['branch'] is not None:
+            branch = self.vals['branch']
+            remote_branch = Branch.open(branch)
+            branch_revno = remote_branch.revno()
+        else:
+            branch = self.vals['trunk_branch']
+            branch_revno = None
+        cmd.append('--public-branch=%s'  % branch)
+        if branch_revno is not None:
+            cmd.append('--public-branch-revno=%d' % branch_revno)
+
+        # Add any additional options for ec2test-remote.py
+        cmd.extend(self.get_remote_test_options())
+        self.log(
+            'Running tests... (output is available on '
+            'http://%s/)\n' % self._instance.hostname)
+
+        # Try opening a browser pointed at the current test results.
+        if self.open_browser:
+            try:
+                import webbrowser
+            except ImportError:
+                self.log("Could not open web browser due to ImportError.")
+            else:
+                status = webbrowser.open(self._instance.hostname)
+                if not status:
+                    self.log("Could not open web browser.")
+
+        # Run the remote script!  Our execution will block here until the
+        # remote side disconnects from the terminal.
+        self._instance.perform(' '.join(cmd))
+        self._running = True
+
+        if not self.headless:
+            sftp = self._instance.ssh.open_sftp()
+            # We ran to completion locally, so we'll be in charge of shutting
+            # down the instance, in case the user has requested a postmortem.
+            #
+            # We only have 60 seconds to do this before the remote test
+            # script shuts the server down automatically.
+            self._instance.perform(
+                'kill `cat /var/launchpad/ec2test-remote.pid`')
+
+            # deliver results as requested
+            if self.file:
+                self.log(
+                    'Writing abridged test results to %s.\n' % self.file)
+                sftp.get('/var/www/summary.log', self.file)
+            sftp.close()
+        # close ssh connection
+        self._instance.ssh.close()
+
+    def get_remote_test_options(self):
+        """Return the test command that will be passed to ec2test-remote.py.
+
+        Returns a tuple of command-line options and switches.
+        """
+        if '--jscheck' in self.test_options:
+            # We want to run the JavaScript test suite.
+            return ('--jscheck',)
+        else:
+            # Run the normal testsuite with our Zope testrunner options.
+            # ec2test-remote.py wants the extra options to be after a double-
+            # dash.
+            return ('--', self.test_options)
+
+
+
+class AcceptAllPolicy:
+    """We accept all unknown host key."""
+
+    # Normally the console output is supposed to contain the Host key
+    # but it doesn't seem to be the case here, so we trust that the host
+    # we are connecting to is the correct one.
+    def missing_host_key(self, client, hostname, key):
+        pass
+
+
+# XXX: JonathanLange 2009-05-31: Strongly considering turning this into a
+# Bazaar plugin -- probably would make the option parsing and validation
+# easier.
+
+if __name__ == '__main__':
+    parser = optparse.OptionParser(
+        usage="%prog [options] [branch]",
+        description=(
+            "Check out a Launchpad branch and run all tests on an Amazon "
+            "EC2 instance."))
+    parser.add_option(
+        '-f', '--file', dest='file', default=None,
+        help=('Store abridged test results in FILE.'))
+    parser.add_option(
+        '-n', '--no-email', dest='no_email', default=False,
+        action='store_true',
+        help=('Do not try to email results.'))
+    parser.add_option(
+        '-e', '--email', action='append', dest='email', default=None,
+        help=('Email address to which results should be mailed.  Defaults to '
+              'the email address from `bzr whoami`. May be supplied multiple '
+              'times. The first supplied email address will be used as the '
+              'From: address.'))
+    parser.add_option(
+        '-o', '--test-options', dest='test_options', default='-vv',
+        help=('Test options to pass to the remote test runner.  Defaults to '
+              "``-o '-vv'``.  For instance, to run specific tests, you might "
+              "use ``-o '-vvt my_test_pattern'``."))
+    parser.add_option(
+        '-b', '--branch', action='append', dest='branches',
+        help=('Branches to include in this run in sourcecode. '
+              'If the argument is only the project name, the trunk will be '
+              'used (e.g., ``-b launchpadlib``).  If you want to use a '
+              'specific branch, if it is on launchpad, you can usually '
+              'simply specify it instead (e.g., '
+              '``-b lp:~username/launchpadlib/branchname``).  If this does '
+              'not appear to work, or if the desired branch is not on '
+              'launchpad, specify the project name and then the branch '
+              'after an equals sign (e.g., '
+              '``-b launchpadlib=lp:~username/launchpadlib/branchname``). '
+              'Branches for multiple projects may be specified with '
+              'multiple instances of this option. '
+              'You may also use this option to specify the branch of launchpad '
+              'into which your branch may be merged.  This defaults to %s. '
+              'Because typically the important branches of launchpad are owned '
+              'by the launchpad-pqm user, you can shorten this to only the '
+              'branch name, if desired, and the launchpad-pqm user will be '
+              'assumed.  For instance, if you specify '
+              '``-b launchpad=db-devel`` then this is equivalent to '
+              '``-b lp:~launchpad-pqm/launchpad/db-devel``, or the even longer'
+              '``-b launchpad=lp:~launchpad-pqm/launchpad/db-devel``.'
+              % (TRUNK_BRANCH,)))
+    parser.add_option(
+        '-t', '--trunk', dest='trunk', default=False,
+        action='store_true',
+        help=('Run the trunk as the branch'))
+    parser.add_option(
+        '-s', '--submit-pqm-message', dest='pqm_message', default=None,
+        help=('A pqm message to submit if the test run is successful.  If '
+              'provided, you will be asked for your GPG passphrase before '
+              'the test run begins.'))
+    parser.add_option(
+        '--pqm-public-location', dest='pqm_public_location', default=None,
+        help=('The public location for the pqm submit, if a pqm message is '
+              'provided (see --submit-pqm-message).  If this is not provided, '
+              'for local branches, bzr configuration is consulted; for '
+              'remote branches, it is assumed that the remote branch *is* '
+              'a public branch.'))
+    parser.add_option(
+        '--pqm-submit-location', dest='pqm_submit_location', default=None,
+        help=('The submit location for the pqm submit, if a pqm message is '
+              'provided (see --submit-pqm-message).  If this option is not '
+              'provided, the script will look for an explicitly specified '
+              'launchpad branch using the -b/--branch option; if that branch '
+              'was specified and is owned by the launchpad-pqm user on '
+              'launchpad, it is used as the pqm submit location. Otherwise, '
+              'for local branches, bzr configuration is consulted; for '
+              'remote branches, it is assumed that the submit branch is %s.'
+              % (TRUNK_BRANCH,)))
+    parser.add_option(
+        '--pqm-email', dest='pqm_email', default=None,
+        help=('Specify the email address of the PQM you are submitting to. '
+              'If the branch is local, then the bzr configuration is '
+              'consulted; for remote branches "Launchpad PQM '
+              '<launchpad@pqm.canonical.com>" is used by default.'))
+    parser.add_option(
+        '-m', '--machine', dest='machine_id', default=None,
+        help=('The AWS machine identifier (AMID) on which to base this run. '
+              'You should typically only have to supply this if you are '
+              'testing new AWS images. Defaults to trying to find the most '
+              'recent one with an approved owner.'))
+    parser.add_option(
+        '-i', '--instance', dest='instance_type',
+        default=DEFAULT_INSTANCE_TYPE,
+        help=('The AWS instance type on which to base this run. '
+              'Available options are %r. Defaults to `%s`.' %
+              (AVAILABLE_INSTANCE_TYPES, DEFAULT_INSTANCE_TYPE)))
+    parser.add_option(
+        '-p', '--postmortem', dest='postmortem', default=False,
+        action='store_true',
+        help=('Drop to interactive prompt after the test and before shutting '
+              'down the instance for postmortem analysis of the EC2 instance '
+              'and/or of this script.'))
+    parser.add_option(
+        '--headless', dest='headless', default=False,
+        action='store_true',
+        help=('After building the instance and test, run the remote tests '
+              'headless.  Cannot be used with postmortem '
+              'or file.'))
+    parser.add_option(
+        '-d', '--debug', dest='debug', default=False,
+        action='store_true',
+        help=('Drop to pdb trace as soon as possible.'))
+    # Use tabs to force a newline in the help text.
+    fake_newline = "\t\t\t\t\t\t\t"
+    parser.add_option(
+        '--demo', action='append', dest='demo_networks',
+        help=("Don't run tests. Instead start a demo instance of Launchpad. "
+              "You can allow multiple networks to access the demo by "
+              "repeating the argument." + fake_newline +
+              "Example: --demo 192.168.1.100 --demo 10.1.13.0/24" +
+              fake_newline +
+              "See" + fake_newline +
+              "https://wiki.canonical.com/Launchpad/EC2Test/ForDemos" ))
+    parser.add_option(
+        '--open-browser', dest='open_browser', default=False,
+        action='store_true',
+        help=('Open the results page in your default browser'))
+    parser.add_option(
+        '-c', '--include-download-cache-changes',
+        dest='include_download_cache_changes', action='store_true',
+        help=('Include any changes in the download cache (added or unknown) '
+              'in the download cache of the test run.  Note that, if you have '
+              'any changes in your download cache, trying to submit to pqm '
+              'will always raise an error.  Also note that, if you have any '
+              'changes in your download cache, you must explicitly choose to '
+              'include or ignore the changes.'))
+    parser.add_option(
+        '-g', '--ignore-download-cache-changes',
+        dest='include_download_cache_changes', action='store_false',
+        help=('Ignore any changes in the download cache (added or unknown) '
+              'in the download cache of the test run.  Note that, if you have '
+              'any changes in your download cache, trying to submit to pqm '
+              'will always raise an error.  Also note that, if you have any '
+              'changes in your download cache, you must explicitly choose to '
+              'include or ignore the changes.'))
+    options, args = parser.parse_args()
+    if options.debug:
+        import pdb; pdb.set_trace()
+    if options.demo_networks:
+        # We need the postmortem console to open the ec2 instance's
+        # network access, and to keep the ec2 instance from being shutdown.
+        options.postmortem = True
+    if len(args) == 1:
+        if options.trunk:
+            parser.error(
+                'Cannot supply both a branch and the --trunk argument.')
+        branch = args[0]
+    elif len(args) > 1:
+        parser.error('Too many arguments.')
+    elif options.trunk:
+        branch = None
+    else:
+        branch = '.'
+    if ((options.postmortem or options.file or options.demo_networks)
+        and options.headless):
+        parser.error(
+            'Headless mode currently does not support postmortem, file '
+            'or demo options.')
+    if options.no_email:
+        if options.email:
+            parser.error(
+                'May not supply both --no-email and an --email address')
+        email = False
+    else:
+        email = options.email
+        if email is None:
+            email = True
+    if options.instance_type not in AVAILABLE_INSTANCE_TYPES:
+        parser.error('Unknown instance type.')
+    if options.branches is None:
+        branches = ()
+    else:
+        branches = [data.split('=', 1) for data in options.branches]
+    runner = EC2TestRunner(
+        branch, email=email, file=options.file,
+        test_options=options.test_options, headless=options.headless,
+        branches=branches,
+        machine_id=options.machine_id, instance_type=options.instance_type,
+        pqm_message=options.pqm_message,
+        pqm_public_location=options.pqm_public_location,
+        pqm_submit_location=options.pqm_submit_location,
+        demo_networks=options.demo_networks,
+        open_browser=options.open_browser, pqm_email=options.pqm_email,
+        include_download_cache_changes=options.include_download_cache_changes,
+        )
+    e = None
+    try:
+        try:
+            runner.start()
+            runner.configure_system()
+            runner.prepare_tests()
+            if options.demo_networks:
+                runner.start_demo_webserver()
+            else:
+                result = runner.run_tests()
+        except Exception, e:
+            # If we are running in demo or postmortem mode, it is really
+            # helpful to see if there are any exceptions before it waits
+            # in the console (in the finally block), and you can't figure
+            # out why it's broken.
+            traceback.print_exc()
+    finally:
+        try:
+            # XXX: JonathanLange 2009-06-02: Blackbox alert! This gets at the
+            # private _instance variable of runner. Instead, it should do
+            # something smarter. For example, the demo networks stuff could be
+            # extracted out to different, non-TestRunner class that has an
+            # instance.
+            if options.demo_networks and runner._instance is not None:
+                demo_network_string = '\n'.join(
+                    '  ' + network for network in options.demo_networks)
+                # XXX: JonathanLange 2009-06-02: Blackbox alert! See above.
+                ec2_ip = socket.gethostbyname(runner._instance.hostname)
+                print (
+                    "\n\n"
+                    "********************** DEMO *************************\n"
+                    "It may take 20 seconds for the demo server to start up."
+                    "\nTo demo to other users, you still need to open up\n"
+                    "network access to the ec2 instance from their IPs by\n"
+                    "entering command like this in the interactive python\n"
+                    "interpreter at the end of the setup. "
+                    "\n  runner.security_group.authorize("
+                    "'tcp', 443, 443, '10.0.0.5/32')\n\n"
+                    "These demo networks have already been granted access on "
+                    "port 80 and 443:\n" + demo_network_string +
+                    "\n\nYou also need to edit your /etc/hosts to point\n"
+                    "launchpad.dev at the ec2 instance's IP like this:\n"
+                    "  " + ec2_ip + "    launchpad.dev\n\n"
+                    "See "
+                    "<https://wiki.canonical.com/Launchpad/EC2Test/ForDemos>."
+                    "\n*****************************************************"
+                    "\n\n")
+             # XXX: JonathanLange 2009-06-02: Blackbox alert! This uses the
+             # private '_instance' variable and assumes that the runner has
+             # exactly one instance.
+            if options.postmortem and runner._instance is not None:
+                console = code.InteractiveConsole({'runner': runner, 'e': e})
+                console.interact((
+                    'Postmortem Console.  EC2 instance is not yet dead.\n'
+                    'It will shut down when you exit this prompt (CTRL-D).\n'
+                    '\n'
+                    'Tab-completion is enabled.'
+                    '\n'
+                    'Test runner instance is available as `runner`.\n'
+                    'Also try these:\n'
+                    '  http://%(dns)s/current_test.log\n'
+                    '  ssh -A %(dns)s') %
+                                 # XXX: JonathanLange 2009-06-02: Blackbox
+                                 # alert! See above.
+                                 {'dns': runner._instance.hostname})
+                print 'Postmortem console closed.'
+        finally:
+            runner.shutdown()

=== added file 'utilities/prioritygrid.py'
--- utilities/prioritygrid.py	1970-01-01 00:00:00 +0000
+++ utilities/prioritygrid.py	2009-08-31 05:15:30 +0000
@@ -0,0 +1,145 @@
+#!/usr/bin/env python
+#
+# Copyright 2009 Canonical Ltd.  This software is licensed under the
+# GNU Affero General Public License version 3 (see the file LICENSE).
+
+# Use this to calculate priorities based on Wiki priority lists such as
+# https://launchpad.canonical.com/VersionThreeDotO/Bugs/Inputs
+
+
+import sys
+
+class Row:
+    def __init__(self, scores, trailing):
+        self.scores = scores
+        self.trailing = trailing
+
+
+def append_average(items, adjusted_scores, row):
+    if len(adjusted_scores) == 0:
+        avg = len(rows_of_scores)
+    else:
+        avg = sum(adjusted_scores)/len(adjusted_scores)
+    items.append((avg, "||%4.1f||%s" % (avg, row.trailing)))
+
+
+def blanks_dont_count(rows_of_scores):
+    items = []
+    for row in rows_of_scores:
+        adjusted_scores = []
+        for score in row.scores:
+            if score == -1:
+                continue
+            adjusted_scores.append(score)
+        append_average(items, adjusted_scores, row)
+    items.sort()
+    return items
+
+
+def blanks_are_heavy(rows_of_scores, half=False):
+    items = []
+    for row in rows_of_scores:
+        adjusted_scores = []
+        for score in row.scores:
+            if score == -1:
+                score = len(rows_of_scores)
+                if half:
+                    score = score/2
+            adjusted_scores.append(score)
+        append_average(items, adjusted_scores, row)
+    items.sort()
+    return items
+
+
+def less_is_more(rows_of_scores):
+    total_per_column = {}
+    for row in rows_of_scores:
+        scores = row.scores
+        for i in range(0, len(scores)):
+            score = scores[i]
+            total_per_column.setdefault(i, 0)
+            if score != -1:
+                total_per_column[i] += 1
+    print total_per_column
+    items = []
+    for row in rows_of_scores:
+        scores = row.scores
+        adjusted_scores = []
+        for i in range(0, len(scores)):
+            score = scores[i]
+            if score == -1:
+                score = len(rows_of_scores)
+            weight = total_per_column[i]
+            adjusted_scores.append(score*weight)
+        append_average(items, adjusted_scores, row)
+    items.sort()
+    return items
+
+
+def condorcet(rows_of_scores):
+    raise NotImplementedError
+
+
+def parse_scores(str):
+    rows = []
+    head = []
+    tail = []
+    ate_first_line = False
+    # We drop the first split element because the line starts with a ||
+    delta = 1
+    for s in str:
+        if not s.strip().startswith("||"):
+            # Regular output; just output it
+            if rows:
+                tail.append(s)
+            else:
+                head.append(s)
+            continue
+
+        if not ate_first_line:
+            ate_first_line = True
+            # Let's take a look at the header
+            if s.strip().startswith("|| * "):
+                # Get rid of scores since we're recalculating
+                delta += 1
+                head.append(s)
+            else:
+                head.append("|| *  " + s)
+            continue
+
+        scores = []
+        columns = s.split("||")[delta:]
+        for col_idx, score in enumerate(columns):
+            score = score.strip()
+            if score:
+                first_char = score[0]
+                if not first_char.isdigit() and not first_char == "-":
+                    # We hit some text, get out of here
+                    break
+            try:
+                score = float(score)
+            except ValueError:
+                # If no value was input, we assume it is equivalent to being
+                # the last option for this voter.
+                score = -1
+            scores.append(score)
+        rows.append(Row(scores, '||'.join(columns)))
+    return rows, head, tail
+
+
+if __name__ == "__main__":
+    str = sys.stdin.read().strip().splitlines()
+    rows_of_scores, head, tail = parse_scores(str)
+    if len(sys.argv) > 1 and sys.argv[1] == "less-is-more":
+        func = less_is_more
+    elif len(sys.argv) > 1 and sys.argv[1] == "condorcet":
+        func = condorcet
+    elif len(sys.argv) > 1 and sys.argv[1] == "blanks-dont-count":
+        func = blanks_dont_count
+    else:
+        func = blanks_are_heavy
+    items = func(rows_of_scores)
+    print "\n".join(head)
+    print "\n".join([i[1] for i in items])
+    print "\n".join(tail)
+

=== added file 'utilities/top-tests.py'
--- utilities/top-tests.py	1970-01-01 00:00:00 +0000
+++ utilities/top-tests.py	2009-08-31 05:15:30 +0000
@@ -0,0 +1,476 @@
+#!/usr/bin/python2.4
+#
+# Copyright 2009 Canonical Ltd.  This software is licensed under the
+# GNU Affero General Public License version 3 (see the file LICENSE).
+
+"""top-tests.py - Report about slowest tests in the test suite.
+
+It parses the output of the testrunner run with -vvv and collects
+statistics about the test run.
+"""
+
+__metaclass__ = type
+
+import re
+import operator
+import os
+import sys
+
+LEN = 20
+
+class ParseException(Exception):
+    """Exception raised when there is an error while parsing a log file."""
+
+
+class TestRunnerStats:
+    """Encapsulates information about the time it took to run a testsuite."""
+
+    LAYER_STARTS_RE = re.compile(r'Running (.+) tests:')
+
+    LAYER_ENDS_RE = re.compile(
+        r'  Ran (\d+) tests with (\d+) failures and (\d+) errors in ([\d.]+) '
+        'seconds.')
+
+    SETUP_RE = re.compile(r'  Set up ([\w.]+) in ([\d.]+) seconds.')
+
+    TEARDOWN_RE = re.compile(r'  Tear down ([\w.]+) in ([\d.]+) seconds.')
+
+    UNSUPPORTED_TEARDOWN_RE = re.compile(
+        r'  Tear down ([\w.]+) ... not supported')
+
+    # We are not restricting this to the standard python identifiers because
+    # some doctest unittest or generated tests could contain funky names.
+    PYTHON_TEST_RE = re.compile(r'([^\( ]+) ?\(([^\)]+)\)')
+
+    MS_RE = re.compile(r'\s*\(([\d.]+) ms\)$')
+
+    TOTAL_RE = re.compile(r'Total: (\d+) tests, (\d+) failures, (\d+) errors')
+
+    # List of strings/patterns to attempt at matching.
+    # The second element in the tuple is the method to call when the start of
+    # the current line matches the string or the pattern.
+    MATCH_LIST = [
+        ('Running tests at level', 'handleStartTestRunner'),
+        (LAYER_STARTS_RE, 'handleLayerStart'),
+        (LAYER_ENDS_RE, 'handleLayerEnd'),
+        (SETUP_RE, 'handleLayerSetUp'),
+        (TEARDOWN_RE, 'handleLayerTearDown'),
+        (UNSUPPORTED_TEARDOWN_RE, 'handleUnsupportedTearDown'),
+        ('  Running:', None),
+        ('Tearing down left over layers:', 'handleFinalTearDown'),
+        (MS_RE, 'handleTestRuntime'),
+        (LAYER_ENDS_RE, 'handleLayerEnd'),
+        (TEARDOWN_RE, 'handleLayerTearDown'),
+        (TOTAL_RE, 'handleTotal'),
+        ('    ', 'handleTestRun'),
+        (None, 'handleGarbage'),
+        ]
+
+    def __init__(self, logfile):
+        """Create a new TestRunnerStats from a log file.
+
+        :param logfile: Open file-like object containing the log of the test
+            suite. That should have been generated at -vvv for maximum
+            information.
+        :raise ParseException: when the log file doesn't contain a testrunner
+            log, or couldn't be parsed for some other reasons.
+        """
+        self.logfile = logfile
+        self._parse()
+
+    def _parse(self):
+        """Extract timing information from the log file."""
+        self.layers = {}
+        self.ignored_lines = []
+        self.current_layer = None
+        self.last_test = None
+
+        end_of_tests = False
+        while not end_of_tests:
+            line = self.logfile.readline()
+            if not line:
+                break
+            for match, action in self.MATCH_LIST:
+                found = False
+                if isinstance(match, basestring):
+                    if line.startswith(match):
+                        found = True
+                elif match is None:
+                    # None indicates the default action.
+                    found = True
+                elif getattr(match, 'match', None):
+                    found = match.match(line)
+                if found:
+                    # Action is the name of the method to call.
+                    # If it returns False, stop parsing.
+                    if action is not None:
+                        end_of_tests = getattr(self, action)(line, found)
+                    break
+
+        if not end_of_tests:
+            raise ParseException('End of file before end of test run.')
+
+    def handleStartTestRunner(self, line, ignored):
+        """Switch the the layer state."""
+
+    def handleLayerStart(self, line, match):
+        """Create a new stats container for the layer."""
+        layer_name = match.group(1)
+        self.current_layer = self.getLayer(layer_name)
+
+    def handleLayerEnd(self, line, match):
+        """Collect the total runtime for the layer tests."""
+        tests_run = match.group(1)
+        runtime = match.group(4)
+        self.current_layer.collectEndResults(tests_run, runtime)
+
+    def handleLayerSetUp(self, line, match):
+        """Collect the runtime for the layer set up."""
+        layer_name = match.group(1)
+        runtime = float(match.group(2))
+        self.getLayer(layer_name).collectSetUp(runtime)
+
+    def handleLayerTearDown(self, line, match):
+        """Collect the runtime for the layer tear down."""
+        layer_name = match.group(1)
+        runtime = float(match.group(2))
+        self.getLayer(layer_name).collectTearDown(runtime)
+
+    def handleUnsupportedTearDown(self, line, match):
+        """Flag that tear down was unsupported."""
+        layer_name = match.group(1)
+        self.getLayer(layer_name).collectUnsupportedTearDown()
+
+    def handleFinalTearDown(self, line, match):
+        """Switch to teardown state."""
+
+    def handleTestRun(self, line, ignored):
+        """Collect that a test was run."""
+        # If we didn't saw the last test runtime, we are probably
+        # in a stack trace or something like that. So treat it as garbage.
+        if self.last_test is not None and not self.last_test_complete:
+            if self.MS_RE.search(line) is None:
+                self.handleGarbage(line, ignored)
+                return
+            else:
+                # It happens that a test doesn't output timing information.
+                # But other tests after that will. 
+                # We are probably encountering such a case.
+                pass
+        line = line[4:]
+        if '/' in line:
+            if ' ' in line:
+                doctest, line = line.split(' ', 1)
+            else:
+                doctest = line
+                line = '\n'
+            self.last_test = DocTestStats(doctest)
+        else:
+            match = self.PYTHON_TEST_RE.match(line)
+            if match:
+                self.last_test = PythonTestStats(
+                    match.group(1), match.group(2))
+            else:
+                raise ParseException("can't parse test name: %s" % line)
+            line = line[match.end():]
+        self.current_layer.collectTest(self.last_test)
+
+        # If the runtime isn't on this line, it means that there was output
+        # by the test, so we'll find the runtime info later on.
+        match = self.MS_RE.search(line)
+        if match:
+            self.last_test_complete = True
+            self.last_test.runtime = float(match.group(1))
+        else:
+            self.last_test_complete = False
+            self.last_test.collectGarbage(line)
+
+    def handleGarbage(self, line, ignored):
+        """Save the log output by the test."""
+        if self.last_test is not None:
+            self.last_test.collectGarbage(line)
+        else:
+            self.ignored_lines.append(line)
+
+    def handleTestRuntime(self, line, match):
+        """Collect the broken test runtime."""
+        if self.last_test is not None:
+            self.last_test.runtime = float(match.group(1))
+            self.last_test_complete = True
+        else:
+            self.ignored_lines.append(line)
+
+    def handleTotal(self, line, match):
+        """Action invoked when the final line is encountered."""
+        self.current_layer = None
+        self.last_test = None
+        return True
+
+    def getLayer(self, layer_name):
+        """Return the layer with name.
+
+        Create and return an empty layer if it doesn't exists.
+        """
+        if layer_name not in self.layers:
+            self.layers[layer_name] = TestLayerStats(layer_name)
+        return self.layers[layer_name]
+
+    def getTestsIter(self):
+        """Return an iterator over all tests."""
+        for layer in self.layers.values():
+            for test in layer.tests:
+                yield test
+
+    @property
+    def total_runtime(self):
+        """Number of seconds used to run the whole test suite."""
+        return sum([layer.total_runtime for layer in self.layers.values()])
+
+    @property
+    def tests_count(self):
+        """Number of tests in the test suite."""
+        return sum([len(layer.tests) for layer in self.layers.values()])
+
+
+class TestLayerStats:
+    """Contain all the tests that were run in the layer."""
+
+    name = None
+    unsupported_tear_downs = 0
+
+    tests_runtime = 0
+
+    def __init__(self, name):
+        """Create a new stats container."""
+        self.name = name
+        self.tests = []
+        self.set_ups = []
+        self.tear_downs = []
+
+    @property
+    def total_runtime(self):
+        """Return the runtime (including fixture) in this layer."""
+        return self.tests_runtime + sum(self.set_ups) + sum(self.tear_downs)
+
+    def collectTest(self, test):
+        """Call when a test was run in the layer."""
+        self.tests.append(test)
+
+    def collectEndResults(self, tests_run, runtime):
+        """Called when all the tests in the layer were run."""
+        self.tests_runtime = float(runtime)
+        self.tests_count = int(tests_run)
+
+    def collectSetUp(self, runtime):
+        """Called when the layer was set up."""
+        self.set_ups.append(runtime)
+
+    def collectTearDown(self, runtime):
+        """Called when the layer was torn down."""
+        self.tear_downs.append(runtime)
+
+    def collectUnsupportedTearDown(self):
+        """Called when the layer couldn't be torn down."""
+        self.unsupported_tear_downs += 1
+
+    def __iter__(self):
+        """Return an iterator over the tests run in this layer."""
+        return iter(self.tests)
+
+
+class TestStats:
+    """Base class for a test stats."""
+
+    name = None
+    runtime = 0
+
+    def __init__(self):
+        self._garbage = []
+
+    @property
+    def garbage(self):
+        """Return the garbage output by the test."""
+        return "".join(self._garbage)
+
+    def collectGarbage(self, line):
+        self._garbage.append(line)
+
+
+class PythonTestStats(TestStats):
+    """Stats for a regular python unit test."""
+
+    def __init__(self, method, module):
+        super(PythonTestStats, self).__init__()
+        self.method = method
+        self.module = module
+
+    @property
+    def name(self):
+        """Return the full name of the test."""
+        return "%s.%s" % (self.module, self.method)
+
+
+class DocTestStats(TestStats):
+    """Stats for a doctest."""
+
+    def __init__(self, filename):
+        super(DocTestStats, self).__init__()
+        self.filename = filename
+
+    @property
+    def name(self):
+        """Remove the PQM directory from the name."""
+        index = self.filename.find("lib/canonical")
+        if index != -1:
+            filename = self.filename[index:]
+        else:
+            filename = self.filename
+        return os.path.normpath(filename)
+
+
+class PQMLog:
+    """Encapsulates information about a PQM log."""
+
+    def __init__(self, logfile):
+        """Create a new PQMLog instance.
+
+        :param logfile: Path to the PQM log.
+        """
+        self.logfile = logfile
+        self.fixtures_profile = []
+
+        self._parse()
+
+    def _parse(self):
+        """Parse a PQM log file.
+
+        Extract the branch name, the time each tests took as well as the
+        time spent in the layers.
+        """
+        self.branch = "Unknown"
+        profile = self.fixtures_profile
+
+        logfile = open(self.logfile)
+        while True:
+            line = logfile.readline()
+            if not line:
+                break
+            line = line.strip()
+            if not line:
+                continue
+            if line.startswith("Executing star-merge"):
+                self.branch = line.split(" ")[2]
+            elif " calls taking " in line:
+                if "s." not in line:
+                    continue
+                values = line.split(" ")
+                runtime = float(values[-1][:-2])
+                profile.append((runtime, values[0]))
+            elif line.startswith("Executing pre-commit hook"):
+                self.testrunner_stats = TestRunnerStats(logfile)
+
+
+def main(argv):
+    """Parse a PQM log file."""
+    if len(sys.argv) > 1:
+        logfile = sys.argv[1]
+    else:
+        logfile = find_latest_successful_merge()
+    print_report(PQMLog(logfile))
+
+
+def find_latest_successful_merge():
+    """Return the latest PQM log that contain a successful merge.
+
+    Look into the current directory for the log files.
+    """
+    cmd = "ls -at | head -10 | xargs grep -l 'star-merge succeeded'"
+    p = os.popen(cmd)
+    logfile_name = p.readlines()[0].strip()
+    p.close()
+    return logfile_name
+
+
+def print_report(pqm_log, out=sys.stdout):
+    """Print the report on STDOUT."""
+
+    print >>out, "Log: %s" % pqm_log.logfile
+    print >>out, "Branch: %s" % pqm_log.branch
+
+    stats = pqm_log.testrunner_stats
+
+    top_tests =  list(stats.getTestsIter())
+    top_tests.sort(key=operator.attrgetter('runtime'), reverse=True)
+
+    total_runtime = stats.total_runtime
+    tests_count = stats.tests_count
+
+    print >>out
+    print >>out, "Top %d tests taking the longest time" % LEN
+    print >>out, "===================================="
+    print
+    top_runtime = 0.0
+    for test in top_tests[:LEN]:
+        percent = test.runtime / total_runtime * 100
+        top_runtime += test.runtime
+        print >>out, "%6.2f (%.1f%%) %s" % (test.runtime, percent, test.name)
+    print >>out
+    test_percent = LEN / float(tests_count) * 100
+    time_percent = top_runtime / total_runtime * 100
+    print >>out, (
+        "Top %s of %s (%.1f%%) tests taking %ss of %ss (%.1f%%)"
+        % (LEN, tests_count, test_percent, top_runtime, total_runtime,
+           time_percent))
+    print >>out
+
+    print >>out, "Tests and runtime by layer"
+    print >>out, "=========================="
+    print >>out
+
+    layers = stats.layers.values()
+    layers.sort(key=operator.attrgetter('total_runtime'), reverse=True)
+    for layer in layers:
+        if len(layer.tests) == 0:
+            continue
+        runtime_percent = layer.tests_runtime / total_runtime * 100
+        layer_name = layer.name.split('.')[-1]
+        print "%7.2f (%4.1f%%) %4d tests (%5.2fs/t) %s" % (
+            layer.tests_runtime, runtime_percent, len(layer.tests),
+            layer.tests_runtime / len(layer.tests), layer_name)
+
+
+    print >>out
+    print >>out, "Slowest fixture methods"
+    print >>out, "======================="
+    print >>out
+
+    profile = list(pqm_log.fixtures_profile)
+    profile.sort(reverse=True)
+    print >>out
+    fixture_runtime = 0
+    for runtime, method in profile:
+        runtime_percent = runtime / total_runtime * 100
+        print >>out, "%7.2f (%4.1f%%) %s" % (runtime, runtime_percent, method)
+        fixture_runtime += runtime
+
+    print >>out
+    print >>out, "Fixture overhead %ss (%.1f%%)" % (
+        fixture_runtime, fixture_runtime / total_runtime * 100)
+    print >>out
+
+    tests_with_garbage = 0
+    garbage_lines_count = 0
+    for test in stats.getTestsIter():
+        if len(test.garbage):
+            tests_with_garbage += 1
+            garbage_lines_count += test.garbage.strip().count('\n')+1
+
+    print >>out, "%d tests output %d warning lines." % (
+        tests_with_garbage, garbage_lines_count)
+    print >>out, "Ignored %d lines in the testrunner output." % len(
+        stats.ignored_lines)
+    print >>out
+
+
+if __name__ == '__main__':
+    main(sys.argv)

